"""
Get shipment tracking info from Posti
"""

from __future__ import unicode_literals, print_function, division
from dateutil.parser import parse
from dateutil.tz import tzutc
from datetime import datetime
from urllib.parse import quote_plus


def init(bot):
    global lang
    config = bot.config.get("module_posti", {})
    lang = config.get("language", "en")


def command_posti(bot, user, channel, args):
    """Get latest tracking event for a shipment from Posti. Usage: .posti JJFI00000000000000"""

    if not args:
        return bot.say(channel, "Tracking ID is required.")

    url = (
        "https://www.posti.fi/henkiloasiakkaat/seuranta/api/shipments/%s"
        % quote_plus(args)
    )

    try:
        r = bot.get_url(url)
        r.raise_for_status()
        data = r.json()
        shipment = data["shipments"][0]
    except Exception as e:
        bot.say(
            channel,
            "Error while getting tracking data. Check the tracking ID or try again later.",
        )
        raise e

    phase = shipment["phase"]
    eta_timestamp = shipment.get("estimatedDeliveryTime")
    latest_event = shipment["events"][0]

    dt = datetime.now(tzutc()) - parse(latest_event["timestamp"])

    agestr = []
    if dt.days > 0:
        agestr.append("%dd" % dt.days)
    secs = dt.seconds
    hours, minutes = secs // 3600, secs // 60 % 60
    if hours > 0:
        agestr.append("%dh" % hours)
    if minutes > 0:
        agestr.append("%dm" % minutes)

    ago = "%s %s" % (
        " ".join(agestr),
        {"fi": "sitten", "sv": "sedan", "en": "ago"}[lang],
    )
    description = latest_event["description"][lang]
    location = "%s %s" % (latest_event["locationCode"], latest_event["locationName"])

    msg = " - ".join([ago, description, location])

    if phase != "DELIVERED" and eta_timestamp:
        eta_dt = parse(eta_timestamp)
        eta_txt = eta_dt.strftime("%d.%m.%Y %H:%M")
        msg = "ETA %s - %s" % (eta_txt, msg)

    bot.say(channel, msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=unused-variable,no-member

"""
A modular python bot based on the twisted matrix irc library

@author Riku 'Shrike' Lindblad (shrike@addiktit.net)
@copyright Copyright (c) 2006 Riku Lindblad
@license New-Style BSD
"""

from __future__ import print_function, division
from typing import Dict, Any, Optional, List, Tuple
import sys
import os.path
import time
import fnmatch
import logging
import requests
import logging.handlers
import json
import jsonschema
from copy import deepcopy
import botcore
from util.dictdiffer import DictDiffer
import socket

import colorlogger

USE_COLOR = True

# Make requests quieter by default
requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)

try:
    import yaml
except ImportError:
    print("PyYAML not found, please install from http://pyyaml.org/wiki/PyYAML")
    sys.exit(1)

# twisted imports
try:
    from twisted.internet import reactor, protocol, ssl
except ImportError:
    print(
        "Twisted library not found, please install Twisted from http://twistedmatrix.com/products/download"
    )
    sys.exit(1)

# default timeout for socket connections
socket.setdefaulttimeout(20)


log = logging.getLogger("core")


class Network(object):
    """Represents an IRC network"""

    def __init__(
        self,
        root: str,
        alias: str,
        address: Tuple[str, int],
        nickname: str,
        realname: Optional[str] = None,
        channels: Optional[List[str]] = None,
        linerate: Optional[float] = None,
        password: Optional[str] = None,
        is_ssl: bool = False,
    ) -> None:
        self.root = root
        self.alias = alias  # network name
        self.address = address  # server address
        self.nickname = nickname  # nick to use
        self.realname = realname
        self.channels: List[str] = channels or []  # channels to join
        self.linerate = linerate
        self.password = password
        self.is_ssl = is_ssl

    def __repr__(self):
        return "Network(%r, %r)" % (self.alias, self.address)


class InstantDisconnectProtocol(protocol.Protocol):
    def connectionMade(self):
        self.transport.loseConnection()


class ThrottledClientFactory(protocol.ClientFactory):
    """Client factory that inserts a slight delay to connecting and reconnecting"""

    lostDelay = 10
    failedDelay = 60

    def clientConnectionLost(self, connector, reason):
        print(connector.getDestination())
        log.info(
            "connection lost (%s): reconnecting in %d seconds"
            % (reason, self.lostDelay)
        )
        reactor.callLater(self.lostDelay, connector.connect)

    def clientConnectionFailed(self, connector, reason):
        log.info(
            "connection failed (%s): reconnecting in %d seconds"
            % (reason, self.failedDelay)
        )
        reactor.callLater(self.failedDelay, connector.connect)


class PyFiBotFactory(ThrottledClientFactory):
    """python.fi bot factory"""

    version = "2013-02-19"
    protocol = botcore.PyFiBot
    allBots = None
    moduledir = os.path.join(sys.path[0], "modules/")
    startTime = None
    config = None

    def __init__(self, config: Dict[str, Any]) -> None:
        """Initialize the factory"""
        self.config = config
        self.data: Dict[str, Any] = {}
        self.data["networks"] = {}
        self.ns: Dict[str, Any] = {}

    def startFactory(self):
        self.allBots = {}
        self.starttime = time.time()
        self._loadmodules()
        ThrottledClientFactory.startFactory(self)
        log.info("factory started")

    def stopFactory(self):
        del self.allBots
        ThrottledClientFactory.stopFactory(self)
        log.info("factory stopped")

    def buildProtocol(self, address):
        # we are connecting to a server, don't know which yet
        log.info("Building protocol for %s", address)

        # Go through all defined networks
        for network, server in self.data["networks"].items():
            log.debug("Looking for matching network: %s - %s", server, address)
            # get all of the ipv4 and ipv6 addresses configured for this domain name
            addrinfo = socket.getaddrinfo(server.address[0], server.address[1])
            ips = set()
            for ip in addrinfo:
                ips.add(ip[4][0])  # (2, 1, 6, '', ('192.168.191.241', 6667))

            # if the address we are connecting to matches one of the IPs defined for
            # this network, connect to it and stop looking
            if address.host in ips:
                log.debug("Connecting to %s / %s", server, address)
                p = self.protocol(self.config, server)
                self.allBots[server.alias] = p
                p.factory = self
                return p

        # No address found
        log.error("Unknown network address: " + repr(address))
        return InstantDisconnectProtocol()

    def createNetwork(
        self,
        address,
        alias,
        nickname,
        realname,
        channels=None,
        linerate=None,
        password=None,
        is_ssl=False,
    ):
        self.setNetwork(
            Network(
                "data",
                alias,
                address,
                nickname,
                realname,
                channels,
                linerate,
                password,
                is_ssl,
            )
        )

    def setNetwork(self, net):
        self.data["networks"][net.alias] = net

    def clientConnectionLost(self, connector, reason):
        """Connection lost for some reason"""
        log.info(
            "connection to %s lost: %s" % (str(connector.getDestination().host), reason)
        )

        # find bot that connects to the address that just disconnected
        for n in self.data["networks"].values():
            dest = connector.getDestination()
            if (dest.host, dest.port) == n.address:
                if n.alias in self.allBots:
                    # did we quit intentionally?
                    if not self.allBots[n.alias].hasQuit:
                        # nope, reconnect
                        ThrottledClientFactory.clientConnectionLost(
                            self, connector, reason
                        )
                    del self.allBots[n.alias]
                    return
                else:
                    log.info("No active connection to known network %s" % n.address[0])

    def _finalize_modules(self, modules=None):
        """Call all module finalizers"""
        if modules is None:
            modules = self._findmodules()
        for module in modules:
            # if rehashing (module already in namespace), finalize the old instance first
            if module in self.ns:
                if "finalize" in self.ns[module][0]:
                    log.info("finalize - %s" % module)
                    self.ns[module][0]["finalize"]()

    def _loadmodules(self):
        """Load all modules"""
        self._finalize_modules()
        for module in self._findmodules():
            env = self._getGlobals()
            log.info("load module - %s" % module)
            # Load new version of the module
            with open(os.path.join(self.moduledir, module), "r") as f:
                exec(f.read(), env, env)
            # Initialize module
            if "init" in env:
                log.info("initialize module - %s" % module)
                env["init"](self)
            # Add to namespace so we can find it later
            self.ns[module] = (env, env)

    def _unload_removed_modules(self):
        """Unload modules removed from modules -directory"""
        # find all modules in namespace, which aren't present in modules -directory
        removed_modules = [m for m in self.ns if m not in self._findmodules()]
        self._finalize_modules(removed_modules)

    def _findmodules(self):
        """Find all modules"""
        modules = [
            m
            for m in os.listdir(self.moduledir)
            if m.startswith("module_") and m.endswith(".py")
        ]
        return modules

    def _getGlobals(self):
        """Global methods for modules"""
        g = {}

        # This crap don't work at all, move it to botcore and proxy it to here
        # g["getUrl"] = self.get_url
        # g["get_url"] = self.get_url
        # g["getNick"] = self.getNick
        # g["getIdent"] = self.getIdent
        # g["getHost"] = self.getHost
        # g["isAdmin"] = self.isAdmin
        # g["is_admin"] = self.is_admin
        # g["to_utf8"] = self.to_utf8
        # g["to_unicode"] = self.to_unicode
        return g

    def get_url(self, url, nocache=False, params=None, headers=None, cookies=None):
        browser = "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:49.0) Gecko/20100101 Firefox/49.0"
        # Common session for all requests
        s = requests.session()
        s.stream = True  # Don't fetch content unless asked
        s.headers.update({"User-Agent": browser})
        s.headers.update({"Accept-Language": "*"})
        s.headers.update(
            {
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"
            }
        )
        # Custom headers from requester
        if headers:
            s.headers.update(headers)
        # Custom cookies from requester
        if cookies:
            s.cookies.update(cookies)

        try:
            r = s.get(url, params=params, timeout=5)
        except requests.exceptions.InvalidSchema:
            log.error("Invalid schema in URI: %s" % url)
            return None
        except requests.exceptions.SSLError:
            log.error("SSL Error when connecting to %s" % url)
            return None
        except requests.exceptions.ConnectionError:
            log.error("Connection error when connecting to %s" % url)
            return None

        size = int(r.headers.get("Content-Length", 0)) // 1024
        size = size / 1024 / 1024  # Size in MB

        content_type = r.headers.get("content-type", None)
        if not content_type:
            content_type = "Unknown"

        if size > 5:
            bot.say(
                channel,
                "File size: %s MB - Content-Type: %s" % (int(size), content_type),
            )

        if size > 2:
            log.warn("Content too large, will not fetch: %skB %s" % (size, url))
            return None

        return r

    def getIdent(self, user):
        """Parses ident from nick!user@host
        @type user: string
        @param user: nick!user@host
        @return: ident"""
        return user.split("!", 1)[1].split("@")[0]

    def getHost(self, user):
        """Parses host from nick!user@host
        @type user: string
        @param user: nick!user@host
        @return: host"""
        return user.split("@", 1)[1]

    def isAdmin(self, user: str) -> bool:
        """Check if an user has admin privileges.
        @return: True or False"""
        for pattern in self.config["admins"]:
            if fnmatch.fnmatch(user, pattern):
                return True
        return False

    def to_utf8(self, _string: Any) -> bytes:
        """Convert string to UTF-8 bytes if it is a string"""
        if isinstance(_string, str):
            _string = _string.encode("UTF-8")
        return _string

    def to_unicode(self, _string: Any) -> str:
        """Convert bytes to unicode string"""
        if isinstance(_string, bytes):
            try:
                _string = _string.decode("utf-8")
            except UnicodeDecodeError:
                try:
                    _string = _string.decode("iso-8859-1")
                except UnicodeDecodeError:
                    _string = str(_string)
        elif not isinstance(_string, str):
            _string = str(_string)
        return _string

    def reload_config(self):
        """Reload config-file while bot is running (on rehash)"""
        log = logging.getLogger("reload_config")

        config = read_config()
        if not config:
            return

        valid_config = validate_config(config)
        if not valid_config:
            log.info("Invalid config file!")
            return

        log.info("Valid config file found, reloading...")
        # ignore nick and networks, as we don't want rehash to change these values
        ignored = ["nick", "networks"]
        # make a deep copy of old config, so we don't remove values from it
        old_config = deepcopy(self.config)
        # remove ignored values to make comparing/updating easier and safer
        for k in ignored:
            old_config.pop(k, {})
            config.pop(k, {})

        # Get diff between configs
        dd = DictDiffer(config, old_config)

        for k in dd.added():
            log.info("%s added (%s: %s" % (k, k, config[k]))
            self.config[k] = config[k]
        for k in dd.removed():
            log.info("%s removed (%s: %s)" % (k, k, old_config[k]))
            del self.config[k]
        for k in dd.changed():
            log.info("%s changed" % k)
            if not isinstance(config[k], dict):
                continue

            # compare configs
            d = DictDiffer(config[k], old_config[k])
            # add all changes to a big list
            changes = list(d.added())
            changes.extend(list(d.removed()))
            changes.extend(list(d.changed()))
            # loop through changes and log them individually
            for x in changes:
                log.info(
                    "%s['%s']: '%s' -> '%s'"
                    % (k, x, old_config[k].get(x, {}), config[k].get(x, {}))
                )
            # replace the whole object
            self.config[k] = config[k]

        # change logging level, default to INFO
        log_level = config.get("logging", {}).get("debug", False)
        if log_level:
            logging.root.setLevel(logging.DEBUG)
        else:
            logging.root.setLevel(logging.INFO)

        # change cmdchar, default to "."
        cmdchar = config.get("cmdchar", ".")
        for key, bot in self.allBots.items():
            bot.cmdchar = cmdchar

    def find_bot_for_network(self, network):
        if network not in self.allBots:
            return None
        return self.allBots[network]


def init_logging(config):
    logger = logging.getLogger()

    if config.get("debug", False):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if USE_COLOR:
        FORMAT = (
            "[%(asctime)-15s][%(levelname)-20s][$BOLD%(name)-15s$RESET]  %(message)s"
        )
        # Append file name + number if debug is enabled
        if config.get("debug", False):
            FORMAT = "%s %s" % (FORMAT, " ($BOLD%(filename)s$RESET:%(lineno)d)")
        COLOR_FORMAT = colorlogger.formatter_message(FORMAT, True)
        formatter = colorlogger.ColoredFormatter(COLOR_FORMAT)
    else:
        FORMAT = "%(asctime)-15s %(levelname)-8s %(name)-11s %(message)s"
        formatter = logging.Formatter(FORMAT)
        # Append file name + number if debug is enabled
        if config.get("debug", False):
            FORMAT = "%s %s" % (FORMAT, " (%(filename)s:%(lineno)d)")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def read_config():
    config_file = sys.argv[1] or os.path.join(sys.path[0], "config.yml")

    if os.path.exists(config_file):
        with open(config_file, "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)
    else:
        print(
            "No config file found, please edit example.yml and rename it to config.yml"
        )
        return
    return config


def validate_config(config):
    with open(os.path.join(sys.path[0], "config_schema.json"), "r") as f:
        schema = json.load(f)
    log.info("Validating configuration")
    v = jsonschema.Draft3Validator(schema)
    if not v.is_valid(config):
        log.error("Error(s) in configuration:")
        for error in sorted(v.iter_errors(config), key=str):
            log.error(error)
        return False
    log.info("Config ok")
    return True


def main():
    sys.path.append(os.path.join(sys.path[0], "lib"))

    config = read_config()
    # if config not found or can't validate it, exit with error
    if not config or not validate_config(config):
        sys.exit(1)

    init_logging(config.get("logging", {}))

    factory = PyFiBotFactory(config)
    for network, settings in config["networks"].items():
        # settings = per network, config = global
        nick = settings.get("nick", None) or config["nick"]
        realname = settings.get("realname") or config.get("realname")
        linerate = settings.get("linerate", 0.5) or config.get("linerate", 0.5)
        password = settings.get("password", None)
        is_ssl = bool(settings.get("is_ssl", False))
        port = int(settings.get("port", 6667))
        force_ipv6 = bool(settings.get("force_ipv6", False))

        # normalize channel names to prevent internal confusion
        chanlist = []
        # allow bot to connect even if no channels are declared
        if "channels" in settings:
            for channel in settings["channels"]:
                if channel[0] not in "&#!+":
                    channel = "#" + channel
                chanlist.append(channel)
        else:
            log.warning('No channels defined for "%s"' % network)

        if force_ipv6:
            try:
                server_name = socket.getaddrinfo(
                    settings["server"], port, socket.AF_INET6
                )[0][4][0]
            except IndexError:
                log.error(
                    "No IPv6 address found for %s (force_ipv6 = true)" % (network)
                )
                continue
        else:
            server_name = settings["server"]

        factory.createNetwork(
            (server_name, port),
            network,
            nick,
            realname,
            chanlist,
            linerate,
            password,
            is_ssl,
        )
        if is_ssl:
            log.info("connecting via SSL to %s:%d" % (server_name, port))
            reactor.connectSSL(server_name, port, factory, ssl.ClientContextFactory())
        else:
            log.info("connecting to %s:%d" % (server_name, port))
            reactor.connectTCP(server_name, port, factory)
    reactor.run()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# from tests import bot_mock
# from pyfibot.modules import module_openweather
# from utils import check_re
# import pytest
# from vcr import VCR
# my_vcr = VCR(path_transformer=VCR.ensure_suffix('.yaml'),
#              cassette_library_dir="tests/cassettes/",
#              filter_query_parameters=['appid'])  # censor appid from query


# @pytest.fixture
# def botmock():
#     bot = bot_mock.BotMock()
#     module_openweather.init(bot)
#     return bot


# @my_vcr.use_cassette
# def test_weather(botmock):
#     regex = u'(Lappeenranta, (Finland|FI): Temperature: -?\d+.\d\xb0C, feels like: -?\d+.\d\xb0C, wind: \d+.\d m/s, humidity: \d+%, pressure: \d+ hPa, cloudiness: \d+%)|(Error: API error.)'
#     check_re(regex, module_openweather.command_weather(botmock, None, "#channel", 'lappeenranta')[1])


# @my_vcr.use_cassette
# def test_forecast(botmock):
#     regex = u'(Lappeenranta, (Finland|FI): tomorrow: -?\d+.\d - -?\d+.\d \xb0C \(.*?\), in 2 days: -?\d+.\d - -?\d+.\d \xb0C \(.*?\), in 3 days: -?\d+.\d - -?\d+.\d \xb0C \(.*?\))|(Error: API error.)'
#     check_re(regex, module_openweather.command_forecast(botmock, None, "#channel", 'lappeenranta')[1])



from __future__ import unicode_literals, print_function, division
import feedparser
import dataset
from twisted.internet.reactor import callLater
from threading import Thread
import twisted.internet.error
import logging


logger = logging.getLogger("module_rss")
DATABASE = None
updater = None
botref = None
config = {}


def init(bot, testing=False):
    """Initialize updater"""
    global DATABASE
    global config
    global botref
    global updater

    if testing:
        DATABASE = dataset.connect("sqlite:///:memory:")
    else:
        DATABASE = dataset.connect("sqlite:///databases/rss.db")

    logger.info("RSS module initialized")
    botref = bot
    config = bot.config.get("rss", {})
    finalize()
    # As there's no signal if this is a rehash or restart
    # update feeds in 30 seconds
    updater = callLater(30, update_feeds)


def finalize():
    """Finalize updater (rehash etc) so we don't leave an updater running"""
    global updater
    logger.info("RSS module finalized")
    if updater:
        try:
            updater.cancel()
        except twisted.internet.error.AlreadyCalled:
            pass
    updater = None


def get_feeds(**kwargs):
    """Get feeds from database"""
    return [
        Feed(f["network"], f["channel"], f["id"])
        for f in list(DATABASE["feeds"].find(**kwargs))
    ]


def find_feed(network, channel, **kwargs):
    """Find specific feed from database"""
    f = DATABASE["feeds"].find_one(network=network, channel=channel, **kwargs)
    if not f:
        return
    return Feed(f["network"], f["channel"], f["id"])


def add_feed(network, channel, url):
    """Add feed to database"""
    f = Feed(network=network, channel=channel, url=url)
    return (f.initialized, f.read())


def remove_feed(network, channel, id):
    """Remove feed from database"""
    f = find_feed(network=network, channel=channel, id=int(id))
    if not f:
        return
    DATABASE["feeds"].delete(id=f.id)
    DATABASE["items_%i" % (f.id)].drop()
    return f


def update_feeds(cancel=True, **kwargs):
    # from time import sleep
    """Update all feeds in the DB"""
    global updater
    logger.info("Updating RSS feeds started")
    for f in get_feeds(**kwargs):
        Thread(target=f.update).start()

    # If we get a cancel, cancel the existing updater
    # and start a new one
    # NOTE: Not sure if needed, as atm cancel isn't used in any command...
    if cancel:
        try:
            updater.cancel()
        except twisted.internet.error.AlreadyCalled:
            pass
        updater = callLater(5 * 60, update_feeds)


def command_rss(bot, user, channel, args):
    commands = ["list", "add", "remove", "latest", "update"]

    args = args.split()
    if not args or args[0] not in commands:
        return bot.say(channel, "rss: valid arguments are [%s]" % (", ".join(commands)))

    command = args[0]
    network = bot.network.alias

    # Get latest feed item from database
    # Not needed? mainly for debugging
    # Possibly useful for checking if feed still exists?
    if command == "latest":
        if len(args) < 2:
            return bot.say(channel, 'syntax: ".rss latest <id from list>"')
        feed = find_feed(network=network, channel=channel, id=int(args[1]))
        if not feed:
            return bot.say(channel, "feed not found, no action taken")
        item = feed.get_latest()
        if not item:
            return bot.say(channel, "no items in feed")
        return bot.say(channel, feed.get_item_str(item))

    # List all feeds for current network && channel
    if command == "list":
        feeds = get_feeds(network=network, channel=channel)
        if not feeds:
            return bot.say(channel, "no feeds set up")

        for f in feeds:
            bot.say(channel, "%02i: %s <%s>" % (f.id, f.name, f.url))
        return

    # Rest of the commands are only for admins
    if not bot.factory.isAdmin(user):
        return bot.say(channel, 'only "latest" and "list" available for non-admins')

    # Add new feed for channel
    if command == "add":
        if len(args) < 2:
            return bot.say(channel, 'syntax: ".rss add url"')
        init, items = add_feed(network, channel, url=args[1])
        if not init:
            return bot.say(channel, "feed already added")
        return bot.say(channel, "feed added with %i items" % len(items))

    # remove feed from channel
    if command == "remove":
        if len(args) < 2:
            return bot.say(channel, 'syntax: ".rss remove <id from list>"')
        feed = remove_feed(network, channel, id=args[1])
        if not feed:
            return bot.say(channel, "feed not found, no action taken")
        return bot.say(channel, 'feed "%s" <%s> removed' % (feed.name, feed.url))

    # If there's no args, update all feeds (even for other networks)
    # If arg exists, try to update the feed...
    if command == "update":
        if len(args) < 2:
            bot.say(channel, "feeds updating")
            update_feeds()
            return
        feed = find_feed(network, channel, id=int(args[1]))
        if not feed:
            return bot.say(channel, "feed not found, no action taken")
        feed.update()
        return


class Feed(object):
    """Feed object to simplify feed handling"""

    def __init__(self, network, channel, id=None, url=None):
        # Not sure if (this complex) init is needed...
        self.id = id
        self.network = network
        self.channel = channel
        self.url = url

        if url:
            self.url = url
        self.initialized = False
        # load feed details from database
        self._get_feed_from_db()

    def __repr__(self):
        return "(%s, %s, %s)" % (self.url, self.channel, self.network)

    def __unicode__(self):
        return "%i - %s" % (self.id, self.url)

    def __init_feed(self):
        """Initialize databases for feed"""
        DATABASE["feeds"].insert(
            {
                "network": self.network,
                "channel": self.channel,
                "url": self.url,
                "name": "",
            }
        )
        # Update feed to match the created
        feed = self._get_feed_from_db()
        # Initialize item-database for feed
        self.__save_item(
            {
                "title": "PLACEHOLDER",
                "link": "https://github.com/lepinkainen/pyfibot/",
                "printed": True,
            }
        )
        self.initialized = True
        return feed

    def __get_items_tbl(self):
        """Get table for feeds items"""
        return DATABASE[("items_%i" % (self.id))]

    def __parse_feed(self):
        """Parse items from feed"""
        f = feedparser.parse(self.url)
        if self.initialized:
            self.update_feed_info({"name": f["channel"]["title"]})
        items = [{"title": i["title"], "link": i["link"]} for i in f["items"]]
        return (f, items)

    def __save_item(self, item, table=None):
        """Save item to feeds database"""
        if table is None:
            table = self.__get_items_tbl()
        # If override is set or the item cannot be found, it's a new one
        if not table.find_one(title=item["title"], link=item["link"]):
            # If printed isn't set, set it to the value in self.initialized (True, if initializing, else False)
            # This is to prevent flooding when adding a new feed...
            if "printed" not in item:
                item["printed"] = self.initialized
            table.insert(item)

    def __mark_printed(self, item, table=None):
        """Mark item as printed"""
        if table is None:
            table = self.__get_items_tbl()
        table.update({"id": item["id"], "printed": True}, ["id"])

    def _get_feed_from_db(self):
        """Get self from database"""
        feed = None
        if self.url and not self.id:
            feed = DATABASE["feeds"].find_one(
                network=self.network, channel=self.channel, url=self.url
            )
        if self.id:
            feed = DATABASE["feeds"].find_one(
                network=self.network, channel=self.channel, id=self.id
            )
        if not feed:
            feed = self.__init_feed()
        self.id = feed["id"]
        self.network = feed["network"]
        self.channel = feed["channel"]
        self.url = feed["url"]
        # TODO: Name could just be the domain part of url?
        self.name = feed["name"]
        return feed

    def get_item_str(self, item):
        return "[%s] %s <%s>" % (
            "".join([c for c in self.name][0:18]),
            item["title"],
            item["link"],
        )

    def get_latest(self):
        tbl = self.__get_items_tbl()
        items = [i for i in list(tbl.find(order_by="id"))]
        if not items:
            return
        return items[-1]

    def update_feed_info(self, data):
        """Update feed information"""
        data["id"] = self.id
        if "url" in data:
            self.url = data["url"]
        DATABASE["feeds"].update(data, ["id"])
        # Update self to match new...
        self._get_feed_from_db()

    def read(self):
        """Read new items from feed"""
        f, items = self.__parse_feed()
        # Get table -reference to speed up stuff...
        tbl = self.__get_items_tbl()
        # Save items in DB, saving takes care of duplicate checks
        for i in reversed(items):
            self.__save_item(i, tbl)
        # Set initialized to False, as we have read everything...
        self.initialized = False
        return items

    def get_new_items(self, mark_printed=False):
        """Get all items which are not marked as printed, if mark_printed is set, update printed also."""
        tbl = self.__get_items_tbl()
        items = [i for i in list(tbl.find(printed=False))]
        if mark_printed:
            for i in items:
                self.__mark_printed(i, tbl)
        return items

    def update(self):
        # If botref isn't defined, bot isn't running, no need to run
        # (used for tests?)
        if not botref:
            return

        # Read all items for feed
        logger.debug('Feed "%s" updating' % (self.name))
        self.read()
        # Get number of unprinted items (and don't mark as printed)
        items = self.get_new_items(False)

        if len(items) == 0:
            logger.debug(
                'Feed "%s" containes no new items, doing nothing.' % (self.name)
            )
            return

        logger.debug('Feed "%s" updated with %i new items' % (self.name, len(items)))
        # If bot instance isn't found, don't print anything
        bot_instance = botref.find_bot_for_network(self.network)
        if not bot_instance:
            logger.error('Bot instance for "%s" not found, not printing' % (self.name))
            return

        logger.debug('Printing new items for "%s"' % (self.name))
        # Get all new (not printed) items and print them
        items = self.get_new_items(True)
        for i in items:
            bot_instance.say(self.channel, self.get_item_str(i))


if __name__ == "__main__":
    f = Feed(
        "ircnet", "#pyfibot", "http://feeds.feedburner.com/ampparit-kaikki?format=xml"
    )
    f.read()
    for i in f.get_new_items(True):
        print(i)



# -*- coding: utf-8 -*-
"""
Simple unit tests for URL title handler functionality.
Tests key functions without requiring full bot setup.
"""
import pytest
from unittest.mock import Mock, patch
from datetime import datetime

from pyfibot.modules import module_urltitle


def test_get_length_str():
    """Test conversion of seconds to human readable format"""
    # Access the private function directly
    get_length_str = module_urltitle.__get_length_str

    assert get_length_str(0) == "0s"
    assert get_length_str(30) == "30s"
    assert get_length_str(60) == "1m"
    assert get_length_str(90) == "1m30s"
    assert get_length_str(3600) == "1h"
    assert get_length_str(3661) == "1h1m1s"
    assert get_length_str(7200) == "2h"


def test_get_views():
    """Test conversion of view counts to human readable format"""
    get_views = module_urltitle.__get_views

    assert get_views(0) == "0"
    assert get_views(500) == "500"
    assert get_views(1500) == "2k"
    assert get_views(1000000) == "1M"
    assert get_views(1500000000) == "2Billion"


def test_levenshtein_distance():
    """Test Levenshtein distance calculation"""
    distance = module_urltitle._levenshtein_distance

    assert distance("", "") == 0
    assert distance("a", "") == 1
    assert distance("", "a") == 1
    assert distance("abc", "abc") == 0
    assert distance("abc", "ab") == 1
    assert distance("abc", "def") == 3
    assert distance("kitten", "sitting") == 3


def test_escaped_fragment():
    """Test Google escaped fragment URL transformation"""
    escaped_fragment = module_urltitle.__escaped_fragment

    # URL without fragment
    url1 = "http://example.com/page"
    assert escaped_fragment(url1) == url1

    # URL with hash-bang fragment
    url2 = "http://example.com/page#!/content"
    expected2 = "http://example.com/page?_escaped_fragment_=/content"
    assert escaped_fragment(url2) == expected2

    # URL with query and fragment
    url3 = "http://example.com/page?param=value#!/content"
    expected3 = "http://example.com/page?param=value&_escaped_fragment_=/content"
    assert escaped_fragment(url3) == expected3


def test_check_redundant():
    """Test redundant URL/title detection"""
    check_redundant = module_urltitle._check_redundant

    # Very similar URL and title should be redundant
    url1 = "http://example.com/test-page-name"
    title1 = "Test Page Name"
    assert check_redundant(url1, title1) == True

    # Different URL and title should not be redundant
    url2 = "http://example.com/xyz123"
    title2 = "Completely Different Article About Something Else"
    assert check_redundant(url2, title2) == False

    # Short titles with similar content should be redundant
    url3 = "http://test.com/abc"
    title3 = "abc"
    assert check_redundant(url3, title3) == True


class TestUrlTitleCoreLogic:
    """Test core URL title extraction logic"""

    def create_mock_bot(self):
        """Create a minimal mock bot for testing"""
        bot = Mock()
        bot.config = {"module_urltitle": {}}
        return bot

    def test_title_function_basic(self):
        """Test the _title function that formats and outputs titles"""
        bot = self.create_mock_bot()
        bot.say.return_value = ("channel", "message")

        # Test basic title output
        result = module_urltitle._title(bot, "#test", "Test Title")
        bot.say.assert_called_with("#test", "Title: Test Title")

        # Test title with custom prefix
        result = module_urltitle._title(bot, "#test", "Test Title", prefix="Custom:")
        bot.say.assert_called_with("#test", "Custom: Test Title")

    def test_title_function_with_info(self):
        """Test the _title function with additional info tuple"""
        bot = self.create_mock_bot()
        bot.say.return_value = ("channel", "message")

        # Test title with additional info
        title_tuple = ("Test Title", "Extra Info")
        result = module_urltitle._title(bot, "#test", title_tuple)
        bot.say.assert_called_with("#test", "Title: Test Title [Extra Info]")

    def test_title_function_long_title(self):
        """Test title truncation for very long titles"""
        bot = self.create_mock_bot()
        bot.say.return_value = ("channel", "message")

        # Test very long title gets truncated
        long_title = "A" * 500
        result = module_urltitle._title(bot, "#test", long_title)

        # Should be called with truncated title
        call_args = bot.say.call_args[0][1]
        assert len(call_args) <= 410  # "Title: " + 400 chars + "..."
        assert call_args.endswith("...")

    def test_title_function_caching(self):
        """Test that titles are cached when URL is provided"""
        bot = self.create_mock_bot()
        bot.say.return_value = ("channel", "message")

        # Clear any existing cache
        module_urltitle.cache.clear()

        url = "http://example.com/test"
        title = "Test Title"

        # Call with URL should cache the result
        result = module_urltitle._title(bot, "#test", title, url=url)

        # Check that title was cached
        cached_title = module_urltitle.cache.get(url)
        assert cached_title == title


class TestUrlTitleEndToEnd:
    """End-to-end tests with minimal mocking"""

    def setup_method(self):
        """Set up test environment"""
        # Create minimal bot mock
        self.bot = Mock()
        self.bot.config = {"module_urltitle": {}}
        self.bot.say.return_value = ("#channel", "message")

        # Initialize module
        module_urltitle.init(self.bot)

    def test_handle_url_invalid_inputs(self):
        """Test handle_url with invalid inputs"""
        # None URL
        result = module_urltitle.handle_url(
            self.bot, "user", "#channel", None, "message"
        )
        assert result is None

        # Non-string URL
        result = module_urltitle.handle_url(
            self.bot, "user", "#channel", 123, "message"
        )
        assert result is None

        # Empty URL
        result = module_urltitle.handle_url(self.bot, "user", "#channel", "", "message")
        assert result is None

    def test_handle_url_manual_ignore(self):
        """Test handle_url with manual ignore prefix"""
        url = "http://example.com"
        msg = "- " + url

        result = module_urltitle.handle_url(self.bot, "user", "#channel", url, msg)
        assert result is None

    def test_handle_url_spotify_ignore(self):
        """Test that Spotify URLs are ignored"""
        spotify_urls = [
            "https://open.spotify.com/track/123",
            "https://open.spotify.com/album/456",
            "spotify:track:789",
        ]

        for url in spotify_urls:
            result = module_urltitle.handle_url(self.bot, "user", "#channel", url, url)
            assert result is None



# -*- coding: utf-8 -*-
import pytest
import sys
import os
from typing import Optional

# Add the project root to Python path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

try:
    from pyfibot.modules.module_fmi import WeatherData, WAWA
    FMI_AVAILABLE = True
except ImportError:
    # Skip tests if dependencies are not available
    FMI_AVAILABLE = False
    pytestmark = pytest.mark.skip("FMI module dependencies not available")


@pytest.mark.skipif(not FMI_AVAILABLE, reason="FMI module dependencies not available")
class TestWeatherDataValidation:
    """Test WeatherData Pydantic model validation"""

    def test_valid_weather_data(self):
        """Test creation with valid data"""
        data = WeatherData(
            place="Helsinki",
            temperature=20.5,
            feels_like=22.0,
            wind_speed=3.5,
            humidity=65,
            cloudiness=4,
            weather_description="clear sky"
        )
        
        assert data.place == "Helsinki"
        assert data.temperature == 20.5
        assert data.feels_like == 22.0
        assert data.wind_speed == 3.5
        assert data.humidity == 65
        assert data.cloudiness == 4
        assert data.weather_description == "clear sky"

    def test_minimal_valid_data(self):
        """Test creation with only required field"""
        data = WeatherData(place="Tampere")
        
        assert data.place == "Tampere"
        assert data.temperature is None
        assert data.feels_like is None
        assert data.wind_speed is None
        assert data.humidity is None
        assert data.cloudiness is None
        assert data.weather_description is None

    def test_missing_place_field(self):
        """Test that place field is required"""
        with pytest.raises(ValueError):
            WeatherData()

    def test_humidity_percentage_validation_valid(self):
        """Test humidity validation with valid values"""
        # Valid humidity values
        data = WeatherData(place="Test", humidity=0)
        assert data.humidity == 0
        
        data = WeatherData(place="Test", humidity=50)
        assert data.humidity == 50
        
        data = WeatherData(place="Test", humidity=100)
        assert data.humidity == 100

    def test_humidity_percentage_validation_invalid(self):
        """Test humidity validation with invalid values"""
        # Invalid humidity values should be set to None
        data = WeatherData(place="Test", humidity=150)
        assert data.humidity is None
        
        data = WeatherData(place="Test", humidity=-10)
        assert data.humidity is None

    def test_cloudiness_validation_valid(self):
        """Test cloudiness validation with valid values"""
        # Valid cloudiness values (0-8 scale)
        for cloudiness in range(0, 9):
            data = WeatherData(place="Test", cloudiness=cloudiness)
            assert data.cloudiness == cloudiness

    def test_cloudiness_validation_invalid(self):
        """Test cloudiness validation with invalid values"""
        # Invalid cloudiness values should be set to None
        data = WeatherData(place="Test", cloudiness=150)
        assert data.cloudiness is None
        
        data = WeatherData(place="Test", cloudiness=-1)
        assert data.cloudiness is None

    def test_temperature_edge_cases(self):
        """Test temperature with various edge cases"""
        # Very cold temperature
        data = WeatherData(place="Antarctica", temperature=-89.2)
        assert data.temperature == -89.2
        
        # Very hot temperature
        data = WeatherData(place="Death Valley", temperature=56.7)
        assert data.temperature == 56.7
        
        # Zero temperature
        data = WeatherData(place="Freezing Point", temperature=0.0)
        assert data.temperature == 0.0
        
        # Float precision
        data = WeatherData(place="Precise", temperature=20.123456)
        assert data.temperature == 20.123456

    def test_wind_speed_edge_cases(self):
        """Test wind speed with various edge cases"""
        # No wind
        data = WeatherData(place="Calm", wind_speed=0.0)
        assert data.wind_speed == 0.0
        
        # Strong wind
        data = WeatherData(place="Hurricane", wind_speed=50.0)
        assert data.wind_speed == 50.0
        
        # Very precise wind speed
        data = WeatherData(place="Precise Wind", wind_speed=3.14159)
        assert data.wind_speed == 3.14159

    def test_unicode_place_names(self):
        """Test place names with unicode characters"""
        # Finnish place names with special characters
        data = WeatherData(place="Jyväskylä")
        assert data.place == "Jyväskylä"
        
        data = WeatherData(place="Säkylä")
        assert data.place == "Säkylä"
        
        data = WeatherData(place="Ähtäri")
        assert data.place == "Ähtäri"

    def test_long_place_names(self):
        """Test very long place names"""
        long_name = "Very Long Place Name That Might Cause Issues" * 10
        data = WeatherData(place=long_name)
        assert data.place == long_name

    def test_weather_description_variations(self):
        """Test various weather descriptions"""
        descriptions = [
            "clear sky",
            "few clouds",
            "scattered clouds",
            "broken clouds",
            "shower rain",
            "rain",
            "thunderstorm",
            "snow",
            "mist"
        ]
        
        for desc in descriptions:
            data = WeatherData(place="Test", weather_description=desc)
            assert data.weather_description == desc

    def test_complete_weather_data(self):
        """Test with complete realistic weather data"""
        data = WeatherData(
            place="Helsinki",
            temperature=15.2,
            feels_like=16.1,
            wind_speed=4.2,
            humidity=78,
            cloudiness=6,
            weather_description="overcast clouds"
        )
        
        # Verify all fields are set correctly
        assert data.place == "Helsinki"
        assert data.temperature == 15.2
        assert data.feels_like == 16.1
        assert data.wind_speed == 4.2
        assert data.humidity == 78
        assert data.cloudiness == 6
        assert data.weather_description == "overcast clouds"

    def test_model_serialization(self):
        """Test that the model can be serialized to dict"""
        data = WeatherData(
            place="Test",
            temperature=20.0,
            humidity=50
        )
        
        dict_data = data.model_dump()
        assert isinstance(dict_data, dict)
        assert dict_data["place"] == "Test"
        assert dict_data["temperature"] == 20.0
        assert dict_data["humidity"] == 50

    def test_model_json_serialization(self):
        """Test that the model can be serialized to JSON"""
        data = WeatherData(
            place="Test",
            temperature=20.0,
            humidity=50
        )
        
        json_data = data.model_dump_json()
        assert isinstance(json_data, str)
        assert "Test" in json_data
        assert "20.0" in json_data


@pytest.mark.skipif(not FMI_AVAILABLE, reason="FMI module dependencies not available")
class TestWAWAConstants:
    """Test WAWA weather code constants"""

    def test_wawa_constants_exist(self):
        """Test that WAWA constants are properly defined"""
        assert isinstance(WAWA, dict)
        assert len(WAWA) > 0

    def test_common_wawa_codes(self):
        """Test common WAWA weather codes"""
        # Test some known codes
        expected_codes = {
            10: "utua",
            20: "sumua", 
            21: "sadetta",
            60: "vesisadetta",
            70: "lumisadetta",
            80: "sadekuuroja"
        }
        
        for code, description in expected_codes.items():
            assert code in WAWA
            assert WAWA[code] == description

    def test_wawa_code_types(self):
        """Test that WAWA codes are integers and descriptions are strings"""
        for code, description in WAWA.items():
            assert isinstance(code, int)
            assert isinstance(description, str)
            assert len(description) > 0

    def test_wawa_code_ranges(self):
        """Test that WAWA codes are in expected ranges"""
        for code in WAWA.keys():
            # Weather codes should be positive integers
            assert code > 0
            # Most WAWA codes are in the range 10-99
            assert 10 <= code <= 99

    def test_finnish_weather_descriptions(self):
        """Test that weather descriptions are in Finnish"""
        finnish_words = ["sade", "lumi", "sumu", "utua", "kuuro"]
        descriptions = list(WAWA.values())
        
        # At least some descriptions should contain Finnish weather terms
        finnish_found = any(
            any(word in desc for word in finnish_words)
            for desc in descriptions
        )
        assert finnish_found

    def test_no_duplicate_descriptions(self):
        """Test that there are no duplicate weather descriptions"""
        descriptions = list(WAWA.values())
        unique_descriptions = set(descriptions)
        
        # Allow some duplicates as weather conditions can overlap
        # but there shouldn't be too many
        assert len(unique_descriptions) >= len(descriptions) * 0.7

    def test_precipitation_codes(self):
        """Test specific precipitation-related codes"""
        precipitation_codes = [21, 23, 24, 60, 61, 62, 63, 70, 71, 72, 73]
        
        for code in precipitation_codes:
            if code in WAWA:
                description = WAWA[code]
                # Should contain precipitation-related terms
                precipitation_terms = ["sade", "lumi", "räntä", "jyväs", "kuuro"]
                assert any(term in description for term in precipitation_terms)

    def test_fog_codes(self):
        """Test fog-related codes"""
        fog_codes = [10, 20, 30, 31, 32, 33, 34]
        
        for code in fog_codes:
            if code in WAWA:
                description = WAWA[code]
                # Should contain fog-related terms
                assert "sumu" in description or "utua" in description

    def test_intensity_descriptors(self):
        """Test that intensity descriptors are used"""
        intensity_words = ["heikko", "kohtalainen", "kova", "tiheä"]
        descriptions = list(WAWA.values())
        
        # Some descriptions should have intensity descriptors
        intensity_found = any(
            any(word in desc for word in intensity_words)
            for desc in descriptions
        )
        assert intensity_found


@pytest.mark.skipif(not FMI_AVAILABLE, reason="FMI module dependencies not available") 
class TestRealisticWeatherScenarios:
    """Test realistic weather scenarios"""

    def test_summer_weather(self):
        """Test typical summer weather data"""
        summer_data = WeatherData(
            place="Helsinki",
            temperature=25.3,
            feels_like=27.1,
            wind_speed=2.1,
            humidity=55,
            cloudiness=2,
            weather_description="few clouds"
        )
        
        assert summer_data.temperature > 20
        assert summer_data.humidity < 80
        assert summer_data.cloudiness <= 3

    def test_winter_weather(self):
        """Test typical winter weather data"""
        winter_data = WeatherData(
            place="Rovaniemi",
            temperature=-15.7,
            feels_like=-22.3,
            wind_speed=5.8,
            humidity=85,
            cloudiness=8,
            weather_description="snow"
        )
        
        assert winter_data.temperature < 0
        assert winter_data.feels_like < winter_data.temperature
        assert winter_data.humidity > 70

    def test_rainy_weather(self):
        """Test rainy weather conditions"""
        rainy_data = WeatherData(
            place="Tampere",
            temperature=12.1,
            feels_like=10.8,
            wind_speed=6.2,
            humidity=95,
            cloudiness=8,
            weather_description="heavy rain"
        )
        
        assert rainy_data.humidity >= 90
        assert rainy_data.cloudiness >= 7
        assert "rain" in rainy_data.weather_description.lower()

    def test_extreme_cold(self):
        """Test extreme cold weather"""
        extreme_cold = WeatherData(
            place="Kittilä",
            temperature=-35.2,
            feels_like=-42.1,
            wind_speed=3.5,
            humidity=70,
            cloudiness=1,
            weather_description="clear sky"
        )
        
        assert extreme_cold.temperature < -30
        assert extreme_cold.feels_like < extreme_cold.temperature

    def test_hot_summer_day(self):
        """Test very hot summer conditions"""
        hot_summer = WeatherData(
            place="Turku",
            temperature=32.8,
            feels_like=35.2,
            wind_speed=1.2,
            humidity=40,
            cloudiness=0,
            weather_description="clear sky"
        )
        
        assert hot_summer.temperature > 30
        assert hot_summer.feels_like > hot_summer.temperature
        assert hot_summer.humidity < 50

    def test_stormy_weather(self):
        """Test stormy weather conditions"""
        storm_data = WeatherData(
            place="Vaasa",
            temperature=8.5,
            feels_like=4.2,
            wind_speed=15.8,
            humidity=88,
            cloudiness=8,
            weather_description="thunderstorm"
        )
        
        assert storm_data.wind_speed > 10
        assert storm_data.feels_like < storm_data.temperature
        assert storm_data.cloudiness == 8


# -*- coding: utf-8 -*-
"""
Commands:
.autoop add channel nick!ident@example.org         adds hostmask to channels auto-op list
.autoop remove channel nick!ident@example.org      removes hostmask from channels auto-op list
.autoop list <channel>                             lists ops in current channel or channel
.autoop status                                      lists channels where user is op
.op
"""

from __future__ import unicode_literals, print_function, division
import sqlite3
import re


COMMANDS = ["add", "remove", "list", "status"]
CHANNEL_PREFIXES = "&#!+"


def init(botconfig):
    open_db(True)


def open_db(createTable=False):
    conn = sqlite3.connect("module_autoop.db")
    c = conn.cursor()
    if createTable:
        c.execute("CREATE TABLE IF NOT EXISTS autoops (channel, hostmask, modes);")
        conn.commit()
    return conn, c


def add_op(channel, hostmask, modes="o"):
    if not get_op_status(channel, hostmask):
        conn, c = open_db()
        c.execute("INSERT INTO autoops VALUES (?, ?, ?);", (channel, hostmask, modes))
        conn.commit()
        conn.close()
        return True


def remove_op(channel, hostmask):
    if get_op_status(channel, hostmask):
        conn, c = open_db()
        c.execute(
            "DELETE FROM autoops WHERE channel = ? AND ? GLOB hostmask;",
            (channel, hostmask),
        )
        conn.commit()
        conn.close()
        return True


def get_op_status(channel, hostmask):
    conn, c = open_db()
    c.execute(
        "SELECT * FROM autoops WHERE channel = ? AND ? GLOB hostmask LIMIT 1;",
        (channel, hostmask),
    )
    if c.fetchone():
        retval = True
    else:
        retval = False
    conn.close()
    return retval


def get_user_channels(hostmask):
    conn, c = open_db()
    c.execute("SELECT channel FROM autoops WHERE ? GLOB hostmask;", (hostmask,))
    rows = c.fetchall()
    conn.close()
    return rows


def get_ops(channel):
    conn, c = open_db()
    c.execute("SELECT hostmask FROM autoops WHERE channel = ?;", (channel,))
    rows = c.fetchall()
    conn.close()
    return rows


def check_hostmask(hostmask):
    rg = re.compile(
        ".*?(!).*?(@).*?", re.IGNORECASE | re.DOTALL
    )  # regex for checking hostmask format
    if re.match(rg, hostmask):
        return True
    return False


def op_user(bot, user, channel):
    nick = bot.getNick(user)
    bot.log("auto-opping %s on %s" % (user, channel))
    bot.mode(channel, True, "o", user=nick)


def handle_userJoined(bot, user, channel):
    if bot.is_admin(user) or get_op_status(channel, user):
        op_user(bot, user, channel)


def command_autoop(bot, user, channel, args):
    """Usage: .autoop [add|remove|list|status]"""

    # disallow all non-admin access
    if not bot.is_admin(user):
        return

    if not args:
        return bot.say(channel, "Valid commands are %s" % ", ".join(map(str, COMMANDS)))

    args = args.split()
    command = args[0]
    if command not in COMMANDS:
        return bot.say(
            channel,
            "Invalid command, valid commands are %s" % ", ".join(map(str, COMMANDS)),
        )

    if command == "list":
        if len(args) > 1:
            list_channel = args[1]
        else:
            list_channel = channel
        if list_channel[0] not in CHANNEL_PREFIXES:
            return bot.say(
                channel, 'Channel name must start with one of "%s"' % CHANNEL_PREFIXES
            )

        ops = get_ops(list_channel)
        if ops:
            return bot.say(
                channel,
                "%s ops: %s" % (list_channel, ", ".join(str(i[0]) for i in ops)),
            )
        return bot.say(channel, "%s doesn't have any ops." % list_channel)

    elif command == "status":
        if channel[0] not in CHANNEL_PREFIXES:
            channel_list = get_user_channels(user)
            if channel_list:
                return bot.say(
                    channel,
                    "You're auto-opped on %s"
                    % ", ".join(str(i[0]) for i in channel_list),
                )
            return bot.say(channel, "You're not auto-opped on any channel.")
        else:
            if get_op_status(channel, user):
                return bot.say(channel, "You're op in %s." % channel)
            return bot.say(channel, "You're not an op in %s." % channel)

    else:
        if bot.is_admin(user):
            if len(args) < 3:
                return bot.say(
                    channel, "Command must have channel and hostmask as arguments."
                )

            if not (args[1][0] in CHANNEL_PREFIXES and check_hostmask(args[2])):
                return bot.say(
                    channel,
                    'Channel name must start with one of "%s" and hostmask must be in format nick!ident@hostma.sk'
                    % CHANNEL_PREFIXES,
                )

            if command == "add":
                if add_op(args[1], args[2]):
                    return bot.say(channel, "OP added.")
                return bot.say(channel, "Adding OP failed, is user already auto-opped?")
            elif command == "remove":
                if remove_op(args[1], args[2]):
                    return bot.say(channel, "OP removed.")
                return bot.say(
                    channel, "Removing OP failed, is user the user auto-opped?"
                )
        else:
            return bot.say(channel, "You must be bot admin to use this feature.")


def command_op(bot, user, channel, args):
    """Get op status from bot if you are authorized"""
    if bot.is_admin(user) or get_op_status(channel, user):
        op_user(bot, user, channel)



import os
import pytest

VCR_RECORD_MODE = os.environ.get("VCR_RECORD_MODE", "once")


def pytest_addoption(parser):
    parser.addoption(
        "--vcrmode",
        action="store",
        type=str,
        default=VCR_RECORD_MODE,
        help="set vcr record mode [once, new_episodes, none, all]",
    )


@pytest.fixture
def vcrmode(request):
    return request.config.getoption("--vcrmode")



# -*- coding: utf-8 -*-
"""
Integration tests for URL title functionality using a minimal approach.
Tests the end-to-end URL title extraction without complex bot setup.
"""
import pytest
from unittest.mock import Mock, patch
import requests
from vcr import VCR
import os

from pyfibot.modules import module_urltitle

# VCR setup
VCR_RECORD_MODE = os.environ.get("VCR_RECORD_MODE", "once")
my_vcr = VCR(
    path_transformer=VCR.ensure_suffix(".yaml"),
    cassette_library_dir="tests/cassettes/",
    record_mode=VCR_RECORD_MODE,
)


def test_url_title_basic_mock():
    """Test URL title extraction with a simple mock"""
    # Create a minimal bot mock
    bot = Mock()
    bot.config = {"module_urltitle": {}}
    bot.say.return_value = ("#channel", "message")

    # Mock get_url to return a simple HTML response
    mock_response = Mock()
    mock_response.headers = {"content-type": "text/html; charset=utf-8"}
    mock_response.content = b"""
    <html>
    <head>
        <title>Test Product - Hamam Hand Towel Kitchen Towel Sweet Navy 50x90cm 100% Cotton</title>
        <meta property="og:title" content="Hamam Hand Towel - Navy Blue Cotton Towel">
    </head>
    <body>
        <h1>Product Page</h1>
    </body>
    </html>
    """

    bot.get_url.return_value = mock_response

    # Initialize the module
    module_urltitle.init(bot)

    # Test URL
    url = "https://shop.example.com/products/test-product"

    # Call handle_url
    result = module_urltitle.handle_url(bot, "testuser", "#testchannel", url, url)

    # Verify bot.say was called
    bot.say.assert_called()

    # Check the message
    call_args = bot.say.call_args[0]
    channel = call_args[0]
    message = call_args[1]

    assert channel == "#testchannel"
    assert message.startswith("Title:")
    # Should use OpenGraph title when available
    assert "Hamam Hand Towel - Navy Blue Cotton Towel" in message


def test_url_title_no_og_title():
    """Test URL title extraction when only regular title is available"""
    bot = Mock()
    bot.config = {"module_urltitle": {}}
    bot.say.return_value = ("#channel", "message")

    # Mock get_url to return HTML with only regular title
    mock_response = Mock()
    mock_response.headers = {"content-type": "text/html; charset=utf-8"}
    mock_response.content = b"""
    <html>
    <head>
        <title>Simple Page Title</title>
    </head>
    <body>
        <h1>Content</h1>
    </body>
    </html>
    """

    bot.get_url.return_value = mock_response
    module_urltitle.init(bot)

    url = "https://example.com/simple"
    result = module_urltitle.handle_url(bot, "testuser", "#testchannel", url, url)

    bot.say.assert_called()
    call_args = bot.say.call_args[0]
    message = call_args[1]

    assert "Title: Simple Page Title" == message


@my_vcr.use_cassette
def test_saaren_taika_url_real():
    """Test the actual Saaren Taika URL with VCR recording"""

    # Create a simple requests-based function to get the URL
    def simple_get_url(url, params=None, headers=None, cookies=None):
        session = requests.Session()
        session.headers.update(
            {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
            }
        )
        if headers:
            session.headers.update(headers)
        if cookies:
            session.cookies.update(cookies)

        try:
            response = session.get(url, params=params)
            return response
        except requests.RequestException:
            return None

    # Test the URL directly
    url = "https://shop.saarentaika.com/collections/hamam-pyyhkeet/products/hamam-kasipyyhe-keittiopyyhe-suloinen-navy-koko-50x90cm-100-puuvillaa-paino-125g-saaren-taika"

    response = simple_get_url(url)

    # Basic checks
    assert response is not None
    assert response.status_code == 200
    assert "text/html" in response.headers.get("content-type", "")

    # Check that we got some content
    assert len(response.content) > 1000  # Should be a substantial page

    # Check for expected content in the HTML
    content_str = response.content.decode("utf-8", errors="ignore")

    # Should contain product-related terms
    content_lower = content_str.lower()
    product_terms = ["hamam", "pyyhe", "navy", "puuvilla"]
    found_terms = [term for term in product_terms if term in content_lower]

    print(f"Found product terms: {found_terms}")
    print(f"Response status: {response.status_code}")
    print(f"Content type: {response.headers.get('content-type')}")
    print(f"Content length: {len(response.content)}")

    # Should find at least some product terms
    assert len(found_terms) >= 1, f"Expected to find product terms in content"


def test_url_title_with_actual_extraction():
    """Test URL title extraction using the actual extraction logic"""
    # Use the __get_bs function to test title extraction
    bot = Mock()
    bot.config = {"module_urltitle": {}}

    # Mock response with realistic Shopify content
    mock_response = Mock()
    mock_response.headers = {"content-type": "text/html; charset=utf-8"}
    mock_response.content = b"""
    <!DOCTYPE html>
    <html>
    <head>
        <title>Hamam kasipyyhe / keittiopyyhe Suloinen Navy koko 50x90cm 100% puuvillaa paino 125g - Saaren Taika</title>
        <meta property="og:title" content="Hamam kasipyyhe / keittiopyyhe Suloinen Navy">
        <meta property="og:description" content="Kaunis hamam pyyhe 100% puuvillaa">
    </head>
    <body>
        <h1>Product Title</h1>
    </body>
    </html>
    """

    bot.get_url.return_value = mock_response

    # Test the internal __get_bs function
    module_urltitle.bot = bot
    bs = module_urltitle.__get_bs(bot, "https://example.com")

    assert bs is not None

    # Test title extraction
    og_title = bs.find("meta", {"property": "og:title"})
    regular_title = bs.find("title")

    assert og_title is not None
    assert regular_title is not None
    assert og_title["content"] == "Hamam kasipyyhe / keittiopyyhe Suloinen Navy"
    assert "Hamam kasipyyhe" in regular_title.text


def test_url_validation_edge_cases():
    """Test URL validation and edge cases"""
    bot = Mock()
    bot.config = {"module_urltitle": {}}
    module_urltitle.init(bot)

    # Test None URL
    result = module_urltitle.handle_url(bot, "user", "#channel", None, "message")
    assert result is None

    # Test non-string URL
    result = module_urltitle.handle_url(bot, "user", "#channel", 123, "message")
    assert result is None

    # Test empty URL
    result = module_urltitle.handle_url(bot, "user", "#channel", "", "message")
    assert result is None

    # Test manual ignore
    result = module_urltitle.handle_url(
        bot, "user", "#channel", "http://example.com", "- http://example.com"
    )
    assert result is None

    # Test Spotify ignore
    result = module_urltitle.handle_url(
        bot,
        "user",
        "#channel",
        "https://open.spotify.com/track/123",
        "https://open.spotify.com/track/123",
    )
    assert result is None


if __name__ == "__main__":
    # Run a quick test
    test_url_title_basic_mock()
    print("Basic mock test passed!")



#!/usr/bin/python
# -*- coding: utf-8 -*-
# (c)2004 Timo Reunanen <parker _et_ wolfenstein _dit_ org>

# Thanks to Shrike for helping this :)

# Made by #python.fi ppl for python ppl

"""Find absolute URLs from strings"""

__version__ = "3.0.3"

import re
import string
from typing import List

_countrycodes = [
    "ac",
    "ad",
    "ae",
    "af",
    "ag",
    "ai",
    "al",
    "am",
    "an",
    "ao",
    "aq",
    "ar",
    "as",
    "at",
    "au",
    "aw",
    "ax",
    "az",
    "ba",
    "bb",
    "bd",
    "be",
    "bf",
    "bg",
    "bh",
    "bi",
    "bj",
    "bm",
    "bn",
    "bo",
    "br",
    "bs",
    "bt",
    "bv",
    "bw",
    "by",
    "bz",
    "ca",
    "cc",
    "cd",
    "cf",
    "cg",
    "ch",
    "ci",
    "ck",
    "cl",
    "cm",
    "cn",
    "co",
    "cr",
    "cs",
    "cu",
    "cv",
    "cx",
    "cy",
    "cz",
    "de",
    "dj",
    "dk",
    "dm",
    "do",
    "dz",
    "ec",
    "ee",
    "eg",
    "eh",
    "er",
    "es",
    "et",
    "fi",
    "fj",
    "fk",
    "fm",
    "fo",
    "fr",
    "ga",
    "gb",
    "gd",
    "ge",
    "gf",
    "gg",
    "gh",
    "gi",
    "gl",
    "gm",
    "gn",
    "gp",
    "gq",
    "gr",
    "gs",
    "gt",
    "gu",
    "gw",
    "gy",
    "hk",
    "hm",
    "hn",
    "hr",
    "ht",
    "hu",
    "id",
    "ie",
    "il",
    "im",
    "in",
    "io",
    "iq",
    "ir",
    "is",
    "it",
    "je",
    "jm",
    "jo",
    "jp",
    "ke",
    "kg",
    "kh",
    "ki",
    "km",
    "kn",
    "kp",
    "kr",
    "kw",
    "ky",
    "kz",
    "la",
    "lb",
    "lc",
    "li",
    "lk",
    "lr",
    "ls",
    "lt",
    "lu",
    "lv",
    "ly",
    "ma",
    "mc",
    "md",
    "mg",
    "mh",
    "mk",
    "ml",
    "mm",
    "mn",
    "mo",
    "mp",
    "mq",
    "mr",
    "ms",
    "mt",
    "mu",
    "mv",
    "mw",
    "mx",
    "my",
    "mz",
    "na",
    "nc",
    "ne",
    "nf",
    "ng",
    "ni",
    "nl",
    "no",
    "np",
    "nr",
    "nu",
    "nz",
    "om",
    "pa",
    "pe",
    "pf",
    "pg",
    "ph",
    "pk",
    "pl",
    "pm",
    "pn",
    "pr",
    "ps",
    "pt",
    "pw",
    "py",
    "qa",
    "re",
    "ro",
    "ru",
    "rw",
    "sa",
    "sb",
    "sc",
    "sd",
    "se",
    "sg",
    "sh",
    "si",
    "sj",
    "sk",
    "sl",
    "sm",
    "sn",
    "so",
    "sr",
    "st",
    "sv",
    "sy",
    "sz",
    "tc",
    "td",
    "tf",
    "tg",
    "th",
    "tj",
    "tk",
    "tl",
    "tm",
    "tn",
    "to",
    "tp",
    "tr",
    "tt",
    "tv",
    "tw",
    "tz",
    "ua",
    "ug",
    "uk",
    "um",
    "us",
    "uy",
    "uz",
    "va",
    "vc",
    "ve",
    "vg",
    "vi",
    "vn",
    "vu",
    "wf",
    "ws",
    "ye",
    "yt",
    "yu",
    "za",
    "zm",
    "zw",
    "aero",
    "biz",
    "com",
    "coop",
    "edu",
    "gov",
    "info",
    "int",
    "mil",
    "museum",
    "name",
    "net",
    "org",
    "pro",
    "me",
    "mobi",
    "xxx",
]

_schemes = ["http://", "https://", "ftp://", "svn://"]

# domainlabel   = alphanum | alphanum *( alphanum | "-" ) alphanum
_domainlabel = r"""
(?:
  [a-z0-9]
  |
  (?:
    [a-z0-9]
    [a-z0-9-]*?
    [a-z0-9]
  )
)
"""

# toplabel      = alpha | alpha *( alphanum | "-" ) alphanum
_toplabel = r"""
(?P<toplabel>
  (?:
    [a-z]
    [a-z0-9-]*
    [a-z0-9]
  )
  |
  (?: [a-z] )
)
"""


# hostname      = *( domainlabel "." ) toplabel [ "." ]
_hostname = (
    r"""
(?P<hostname>
  (?: %(_domainlabel)s [.])+
  %(_toplabel)s
)
"""
    % globals()
)

_ipv4 = r"""
(?P<ipv4>
  (?:(?<=[^0-9]) | \A)                  # Doesn't start with number
  (?:[0-9]{1,3}\.){3}[0-9]{1,3}         # '123.123.123.123'
  (?![0-9])                             # And doesn't end with number
)
"""

_singleNum = r"""[0-9a-f]{0,4}"""
_ipv6 = (
    r"""
(?:
  \[                                     # Start [
  (?P<ipv6>
    (?: %(_singleNum)s[:] ){2,7}         # xxxx:xxxx:[xxxx:[xxxx:[xxxx:[xxxx:[xxxx:]]]]]
    %(_singleNum)s                       # xxxx
  )
  \]                                     # End ]
)
"""
    % globals()
)

# host          = hostname | IPv4address | IPv6address [ ":" port ]
_host = (
    r"""
(?P<host>
  (?:                                   # Must be one of these
    %(_hostname)s |
    %(_ipv4)s |
    %(_ipv6)s
  )
  (?:                                   # Check if port is defined
    :
     (?P<port>
      [0-9]+
    )
  )?
)
"""
    % globals()
)

# def emacsHack():
#   pass
# "

validPathChars = [
    string.ascii_letters,
    string.digits,
    "/;?:@&=+$,%#" "-_.!~*()",
    r"ÅÄÖåäö[]<>{}^\|\'`–",  # not so valid but used
]

validUserinfoChars = [string.ascii_letters, string.digits, ";:&=+$,", "-_.!~*()", "%@"]

validPathChars = "".join(validPathChars)
validUserinfoChars = "".join(validUserinfoChars)

hostRe = re.compile(_host, re.VERBOSE | re.MULTILINE | re.I)


def grab(txt: str, needScheme: bool = True) -> List[str]:
    seekpos: int = 0

    possibleUrls: List[str] = []

    # Try to find all possible URI's
    while 1:
        srch = hostRe.search(txt, seekpos)
        if not srch:
            break

        valid = False

        # Find possible URI with valid hostname, ipv4 or ipv6
        if srch.group("host"):
            if srch.group("ipv4"):
                # Check if it is valid ipv4
                ip = [x for x in srch.group("ipv4").split(".") if int(x) > 255]
                if ip == []:
                    valid = True

            elif srch.group("hostname"):
                # Check if valid country code
                if srch.group("toplabel") in _countrycodes:
                    valid = True

            elif srch.group("ipv6"):
                # ipv6 is always valid :P (yeah right...)
                valid = True

        if not valid:
            seekpos = srch.end("host")
            continue

        s = srch.start("host")
        e = srch.end("host")

        # Check if there is userinfo in URI
        if s > 1 and txt[s - 1] == "@":
            s -= 1

            while s >= 0 and txt[s] in validUserinfoChars:
                s -= 1

            s += 1

        # Check if there is scheme
        schemeFound = False
        for scheme in _schemes:
            if txt[s - len(scheme) : s] == scheme:
                s -= len(scheme)
                schemeFound = True
                break

        # If scheme neede and no scheme found, skip it
        if needScheme and not schemeFound:
            seekpos = srch.end("host")
            continue

        # Try to figure out rest of URI
        while e < len(txt) and txt[e] in validPathChars:
            e += 1

        # Some silly hack
        ns = s

        for v in [("<", ">"), ("(", ")"), ("{", "}"), ("[", "]"), '"', "'"]:
            if isinstance(v, tuple):
                sc, ec = v
            else:
                sc = ec = v

            if txt[ns - len(sc) :].startswith(sc):
                if txt[e - len(ec) :].startswith(ec):
                    e -= len(ec)

        possibleUrls.append(txt[s:e])

        seekpos = e

    return possibleUrls



# -*- coding: utf-8 -*-
import pytest
import sys
import os

# Add the project root to Python path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from pyfibot.util.dictdiffer import DictDiffer


class TestDictDiffer:
    """Test the DictDiffer utility class"""

    def test_empty_dicts(self):
        """Test with empty dictionaries"""
        differ = DictDiffer({}, {})
        assert differ.added() == set()
        assert differ.removed() == set()
        assert differ.changed() == set()
        assert differ.unchanged() == set()

    def test_identical_dicts(self):
        """Test with identical dictionaries"""
        dict1 = {"a": 1, "b": 2, "c": 3}
        dict2 = {"a": 1, "b": 2, "c": 3}
        differ = DictDiffer(dict1, dict2)
        
        assert differ.added() == set()
        assert differ.removed() == set()
        assert differ.changed() == set()
        assert differ.unchanged() == {"a", "b", "c"}

    def test_added_keys(self):
        """Test detection of added keys"""
        past_dict = {"a": 1, "b": 2}
        current_dict = {"a": 1, "b": 2, "c": 3, "d": 4}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == {"c", "d"}
        assert differ.removed() == set()
        assert differ.changed() == set()
        assert differ.unchanged() == {"a", "b"}

    def test_removed_keys(self):
        """Test detection of removed keys"""
        past_dict = {"a": 1, "b": 2, "c": 3, "d": 4}
        current_dict = {"a": 1, "b": 2}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == set()
        assert differ.removed() == {"c", "d"}
        assert differ.changed() == set()
        assert differ.unchanged() == {"a", "b"}

    def test_changed_values(self):
        """Test detection of changed values"""
        past_dict = {"a": 1, "b": 2, "c": 3}
        current_dict = {"a": 10, "b": 2, "c": 30}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == set()
        assert differ.removed() == set()
        assert differ.changed() == {"a", "c"}
        assert differ.unchanged() == {"b"}

    def test_complex_scenario(self):
        """Test a complex scenario with all types of changes"""
        past_dict = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}
        current_dict = {"a": 10, "b": 2, "f": 6, "g": 7}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == {"f", "g"}
        assert differ.removed() == {"c", "d", "e"}
        assert differ.changed() == {"a"}
        assert differ.unchanged() == {"b"}

    def test_different_value_types(self):
        """Test with different value types"""
        past_dict = {"str": "hello", "int": 42, "list": [1, 2, 3], "none": None}
        current_dict = {"str": "world", "int": 42, "list": [1, 2, 4], "bool": True}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == {"bool"}
        assert differ.removed() == {"none"}
        assert differ.changed() == {"str", "list"}
        assert differ.unchanged() == {"int"}

    def test_nested_dicts(self):
        """Test with nested dictionaries"""
        past_dict = {"nested": {"a": 1, "b": 2}, "simple": "value"}
        current_dict = {"nested": {"a": 1, "b": 3}, "simple": "value"}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == set()
        assert differ.removed() == set()
        assert differ.changed() == {"nested"}  # The nested dict changed
        assert differ.unchanged() == {"simple"}

    def test_mixed_key_types(self):
        """Test with mixed key types (strings, integers, tuples)"""
        past_dict = {"str_key": 1, 42: "int_key", (1, 2): "tuple_key"}
        current_dict = {"str_key": 2, 42: "int_key", (3, 4): "new_tuple"}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == {(3, 4)}
        assert differ.removed() == {(1, 2)}
        assert differ.changed() == {"str_key"}
        assert differ.unchanged() == {42}

    def test_none_values(self):
        """Test handling of None values"""
        past_dict = {"a": None, "b": 1, "c": None}
        current_dict = {"a": 1, "b": None, "c": None}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == set()
        assert differ.removed() == set()
        assert differ.changed() == {"a", "b"}
        assert differ.unchanged() == {"c"}

    def test_large_dicts(self):
        """Test performance with larger dictionaries"""
        past_dict = {f"key_{i}": i for i in range(1000)}
        current_dict = {f"key_{i}": i * 2 for i in range(500, 1500)}
        differ = DictDiffer(current_dict, past_dict)
        
        added = differ.added()
        removed = differ.removed()
        changed = differ.changed()
        unchanged = differ.unchanged()
        
        # Keys 0-499 were removed
        assert len(removed) == 500
        # Keys 1000-1499 were added
        assert len(added) == 500
        # Keys 500-999 had values changed (i vs i*2)
        assert len(changed) == 500
        # No keys should be unchanged since all overlapping keys had value changes
        assert len(unchanged) == 0

    def test_boolean_values(self):
        """Test with boolean values"""
        past_dict = {"flag1": True, "flag2": False, "flag3": True}
        current_dict = {"flag1": False, "flag2": False, "flag4": True}
        differ = DictDiffer(current_dict, past_dict)
        
        assert differ.added() == {"flag4"}
        assert differ.removed() == {"flag3"}
        assert differ.changed() == {"flag1"}
        assert differ.unchanged() == {"flag2"}


"""Upload images encountered on channels to imgur.com"""

import requests
import logging
import os.path
import sys

log = logging.getLogger("imgur")

# Bot application ID
CLIENT_ID = ""
CLIENT_SECRET = ""

DATAFILE = os.path.join(sys.path[0], "modules", "imgur_auth.dat")

album_id = ""
access_token = ""
refresh_token = ""

init_ok = False


def init(bot):
    try:
        config = config = bot.config.get("module_imgur", {})
    except KeyError:
        config = None

    global album_id
    album_id = config.get("album_id", "")

    # If refreshed tokens exist, get them from the file
    # else load seed values from config
    global access_token, refresh_token
    if os.path.exists(DATAFILE):
        f = open(DATAFILE, "r")
        access_token = f.readline().strip()
        refresh_token = f.readline().strip()
        f.close()
        log.info("Loaded imgur credentials from %s" % DATAFILE)
    else:
        access_token = config.get("access_token", "")
        refresh_token = config.get("refresh_token", "")

    global init_ok
    if access_token and refresh_token:
        log.info("imgur auth tokens found")
        init_ok = True


# Upload all files ending in jpg and gif
def handle_url(bot, user, channel, url, msg):
    if not init_ok:
        return
    # TODO: smarter checking, using content-type perhaps?
    r = bot.get_url(url)
    content_type = r.headers["content-type"].split(";")[0]

    if content_type.startswith("image/"):
        # log.info(channel, upload_images([url]))
        responses = upload_images([url], user, channel)
        for r in responses:
            log.debug(
                "Uploaded image http://imgur.com/%s by %s from %s to gallery"
                % (r.get("id", ""), user, channel)
            )


# Upload a page of images
def upload_gallery(url):
    from mechanize import Browser

    br = Browser()
    br.set_handle_robots(False)
    br.open(url)

    urls = set()

    for link in br.links(url_regex=".jpg$"):
        urls.add(link.url)

    upload_images(urls)


def upload_images(urls, user=None, channel=None):
    global access_token, refresh_token
    # uploaded images
    images = []

    api = "https://api.imgur.com/3"

    headers = {"Authorization": "Bearer %s" % access_token}

    # Transload image(s) to imgur
    for url in urls:
        log.debug("Transloading %s" % url)

        metadata = {
            "image": url,
            "type": "url",
            "album": album_id,
            "title": "by: %s from: %s" % (user, channel),
            "description": "Original url: %s" % (url),
        }

        r = requests.post("%s/upload" % api, headers=headers, data=metadata)

        # error handling
        if r.status_code != 200:
            if (
                r.status_code == 403
                and r.json()["data"]["error"]
                == "The access token provided has expired."
            ):
                access_token, refresh_token = _refresh_token(
                    CLIENT_ID, CLIENT_SECRET, refresh_token
                )
                # recursive retry, kinda dangerous, but what the heck
                return upload_images([url], user, channel)
            else:
                log.error("Transload error for %s: %s " % (url, r.text))
                continue

        images.append(r.json()["data"])

        # log.info("%s uploaded to imgur gallery" % r.json()['data'].get('link', ''))

    return images


def _refresh_token(client_id, client_secret, refresh_token):
    r = requests.post(
        "https://api.imgur.com/oauth2/token",
        data={
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        },
    )

    if r.status_code == 200:
        new_access_token = r.json()["access_token"]
        new_refresh_token = r.json()["refresh_token"]
        log.info(
            "Updated imgur access token for account %s" % r.json()["account_username"]
        )

        f = open(DATAFILE, "w")
        f.write(new_access_token + "\n")
        f.write(new_refresh_token + "\n")
        f.close()

        return new_access_token, new_refresh_token
    else:
        log.error("error refreshing tokens: %s", r.json())
        return None, None



# -*- coding: utf-8 -*-
# pylint: disable=unused-variable,no-member

"""
Bot core

@author Riku 'Shrike' Lindblad (shrike@addiktit.net)
@copyright Copyright (c) 2004 Riku Lindblad
@license New-Style BSD

"""

from __future__ import print_function, division

# twisted imports
from twisted.words.protocols import irc
from twisted.internet import reactor, threads
from twisted.python import rebuild

from types import FunctionType
from typing import Optional, Any, Dict, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from . import pyfibot as pyfibot_module

import inspect
import string
import logging
from util import pyfiurl

# line splitting
import textwrap

__pychecker__ = "unusednames=i, classattr"

log = logging.getLogger("bot")


class CoreCommands(object):
    # Attributes that will be provided by the subclass
    pingAve: float

    def get_nick(self, user: str) -> str:
        """Get nick from user string - implemented in subclass"""
        raise NotImplementedError("get_nick should be implemented by subclass")

    def command_echo(self, user: str, channel: str, args: str) -> None:
        self.say(channel, "%s: %s" % (user, args))

    def command_ping(self, user: str, channel: str, args: str) -> None:
        self.say(
            channel,
            "%s: My current ping is %.0fms"
            % (self.get_nick(user), self.pingAve * 100.0),
        )

    def command_rehash(self, user, channel, args):
        """Reload modules and optionally the configuration file. Usage: rehash [conf]"""

        if self.factory.isAdmin(user):
            try:
                # rebuild core & update
                log.info("rebuilding %r" % self)
                rebuild.updateInstance(self)

                # reload config file
                if args == "conf":
                    self.factory.reload_config()
                    self.say(channel, "Configuration reloaded.")

                # unload removed modules
                self.factory._unload_removed_modules()
                # reload modules
                self.factory._loadmodules()
            except Exception as e:
                self.say(channel, "Rehash error: %s" % e)
                log.error("Rehash error: %s" % e)
            else:
                self.say(channel, "Rehash OK")
                log.info("Rehash OK")

    def say(
        self, channel: str, message: str, length: Optional[int] = None
    ) -> Tuple[str, str, str]:
        """Must be implemented by the inheriting class"""
        raise NotImplementedError

    def command_join(self, user, channel, args):
        """Usage: join <channel>[@network] [password] - Join the specified channel"""

        if not self.factory.isAdmin(user):
            return

        password = None
        # see if we have multiple arguments
        try:
            args, password = args.split(" ", 1)
        except ValueError:
            pass

        # see if the user specified a network
        try:
            newchannel, network = args.split("@", 1)
        except ValueError:
            newchannel, network = args, self.network.alias

        try:
            bot = self.factory.allBots[network]
        except KeyError:
            self.say(channel, "I am not on that network.")
        else:
            log.debug("Attempting to join channel %s on ", (newchannel, network))
            if newchannel in bot.network.channels:
                self.say(channel, "I am already in %s on %s." % (newchannel, network))
                log.debug("Already on channel %s" % channel)
                log.debug("Channels I'm on this network: %s" % bot.network.channels)
            else:
                if password:
                    bot.join(newchannel, key=password)
                    log.debug("Joined with password")
                else:
                    bot.join(newchannel)
                    log.debug("Joined")

    # alias of part
    def command_leave(self, user, channel, args):
        """Usage: leave <channel>[@network] - Leave the specified channel"""
        self.command_part(user, channel, args)

    def command_part(self, user, channel, args):
        """Usage: part <channel>[@network] - Leave the specified channel"""

        if not self.factory.isAdmin(user):
            return

        # part what and where?
        try:
            newchannel, network = args.split("@", 1)
        except ValueError:
            newchannel, network = args, self.network.alias

        # get the bot instance for this chat network
        try:
            bot = self.factory.allBots[network]
        except KeyError:
            self.say(channel, "I am not on that network.")
        else:
            # no arguments, attempt to part current channel
            if not newchannel:
                log.debug("Parted channel %s" % channel)
                bot.network.channels.remove(channel)
                bot.part(channel)
                return

            if newchannel in bot.network.channels:
                log.debug("Parted channel %s" % newchannel)
                bot.network.channels.remove(newchannel)
                bot.part(newchannel)
            else:
                log.debug(
                    "Attempted to part channel I am not on: %s@%s"
                    % (newchannel, network)
                )
                log.debug("Channels on network: %s" % bot.network.channels)

    def command_quit(self, user, channel, args):
        """Usage: logoff - Leave this network"""

        if not self.factory.isAdmin(user):
            return

        self.quit("Working as programmed")
        self.hasQuit = 1

    def command_channels(self, user, channel, args):
        """Usage: channels <network> - List channels the bot is on"""
        if not args:
            self.say(
                channel,
                "Please specify a network: %s" % ", ".join(self.factory.allBots.keys()),
            )
            return

        self.say(channel, "I am on %s" % self.network.channels)

    def command_help(self, user, channel, cmnd):
        """Get help on all commands or a specific one. Usage: help [<command>]"""

        commands, admin_commands = [], []
        for module, env in self.factory.ns.items():
            myglobals, mylocals = env
            commands += [
                (c.replace("command_", ""), ref)
                for c, ref in mylocals.items()
                if c.startswith("command_%s" % cmnd)
            ]
            admin_commands += [
                (c.replace("admin_", ""), ref)
                for c, ref in mylocals.items()
                if c.startswith("admin_%s" % cmnd)
            ]
        # Help for a specific command
        if len(cmnd) > 0:
            for cname, ref in commands:
                if cname == cmnd:
                    helptext = ref.__doc__.split("\n", 1)[0]
                    self.say(channel, "Help for %s: %s" % (cmnd, helptext))
                    return
        # Generic help
        else:
            commandlist = ", ".join([c for c, ref in commands])
            admin_commandlist = ", ".join([c for c, ref in admin_commands])
            self.say(channel, "Available commands: %s" % commandlist)
            if self.factory.isAdmin(user):
                self.say(channel, "Available admin commands: %s" % admin_commandlist)


class PyFiBot(irc.IRCClient, CoreCommands):
    """PyFiBot"""

    nickname = "pyfibot"
    realname: str = "https://github.com/lepinkainen/pyfibot"  # type: ignore[assignment]
    password = None

    # send 2 msgs per second max
    lineRate: float = 0.5  # type: ignore[assignment]
    hasQuit = False

    # Rolling ping time average
    pingAve: float = 0.0

    if TYPE_CHECKING:
        factory: "pyfibot_module.PyFiBotFactory"

    def __init__(self, config: Dict[str, Any], network: Any) -> None:
        self.cmdchar = config.get("cmdchar", ".")
        self.network = network
        self.nickname = self.network.nickname
        self.realname = self.network.realname or self.realname
        self.lineRate = self.network.linerate
        self.password = self.network.password
        # Text wrapper to clip overly long answers
        # 510 is the spec maximum (512 - CRLF)
        self.tw = textwrap.TextWrapper(width=480, break_long_words=True)
        log.info("bot initialized")

    def __repr__(self):
        return "PyFiBot(%r, %r)" % (self.nickname, self.network.address)

    # Core
    def printResult(self, msg, info):
        # Don't print results if there is nothing to say (usually non-operation on module)
        if msg:
            log.debug("Result %s %s" % (msg, info))

    def printError(self, msg, info):
        log.error("ERROR %s %s" % (msg, info))

    def connectionMade(self):
        irc.IRCClient.connectionMade(self)
        self.repeatingPing(300)
        log.info("connection made")

    def connectionLost(self, reason):
        irc.IRCClient.connectionLost(self, reason)
        log.info("connection lost: %s", reason)

    def signedOn(self):
        """Called when bot has succesfully connected to a server."""
        log.info("Connected to network")

        network_conf = self.factory.config["networks"][self.network.alias]

        # Name used for authentication
        authname = network_conf.get("authname", None)
        # Pass used for authentication
        authpass = network_conf.get("authpass", None)

        # If authentication is used
        if authname and authpass:
            # QuakeNet specific auth and IP-address masking
            if self.network.alias.lower() == "quakenet":
                log.info("I'm on Quakenet, authenticating...")
                self.mode(self.nickname, "+", "x")  # Hide ident
                log.info("Authenticating...")
                self.say("Q@CServe.quakenet.org", "AUTH %s %s" % (authname, authpass))
            # more generic authentication
            else:
                # Get authentication service
                # Default: None
                # None as default, so the user must be sure who he's authenticating to
                authservice = network_conf.get("authservice", None)
                if authservice:
                    # Command used for authentication
                    # Default: "IDENTIFY authname authpass"
                    authcommand = network_conf.get(
                        "authcommand", "IDENTIFY %(authname)s %(authpass)s"
                    )
                    authcommand = authcommand % {
                        "authname": authname,
                        "authpass": authpass,
                    }

                    log.info('Authenticating to "%s"' % authservice)
                    log.debug('Authentication command used: "%s"' % authcommand)
                    self.say(authservice, authcommand)
                else:
                    log.info("authservice not set, authentication not attempted")
        else:
            log.debug("authname or authpass not found, authentication not attempted")

        authdelay = network_conf.get("authdelay", None)
        if authdelay:
            # allowing the connection to establish and authentication to happen before joining
            log.info("Joining channels after %s second delay" % (authdelay))
            reactor.callLater(authdelay, self.joinChannels)
        else:
            self.joinChannels()

    # separate function to allow timing the joins
    def joinChannels(self):
        for chan in self.network.channels:
            # Defined as a tuple, channel has a key
            if isinstance(chan, list):
                self.join(chan[0], key=chan[1])
            else:
                self.join(chan)

        log.info(
            "joined %d channel(s): %s"
            % (len(self.network.channels), ", ".join(self.network.channels))
        )
        self._runEvents("signedon")

    def pong(self, user, secs):
        self.pingAve = ((self.pingAve * 5) + secs) / 6.0

    def repeatingPing(self, delay):
        reactor.callLater(delay, self.repeatingPing, delay)
        self.ping(self.nickname)

    # TODO: Move the function here completely
    def get_url(self, url, nocache=False, params=None, headers=None, cookies=None):
        return self.factory.get_url(url, nocache, params, headers, cookies)

    def getUrl(self, url, nocache=False, params=None, headers=None, cookies=None):
        return self.get_url(url, nocache, params, headers, cookies)

    def isAdmin(self, user):
        return self.factory.isAdmin(user)

    def is_admin(self, user):
        return self.factory.isAdmin(user)

    def get_nick(self, user):
        return self.getNick(user)

    def getNick(self, user: str) -> str:
        """Parses nick from nick!user@host
        @type user: string
        @param user: nick!user@host
        @return: nick"""
        return user.split("!", 1)[0]

    def log(self, message):
        botId = "%s@%s" % (self.nickname, self.network.alias)
        log.info("%s: %s", botId, message)

    def callLater(self, delay, callable):
        self.callLater(delay, callable)

    # Communication
    def privmsg(self, user, channel, msg):
        """This will get called when the bot receives a message.
        @param user: nick!user@host
        @param channel: Channel where the message originated from
        @param msg: The actual message
        """

        channel = channel.lower()
        lmsg = msg.lower()
        lnick = self.nickname.lower()
        nickl = len(lnick)
        if channel == lnick:
            # Turn private queries into a format we can understand
            if not msg.startswith(self.cmdchar):
                msg = self.cmdchar + msg
            elif lmsg.startswith(lnick):
                msg = self.cmdchar + msg[nickl:].lstrip()
            elif (
                lmsg.startswith(lnick)
                and len(lmsg) > nickl
                and lmsg[nickl] in string.punctuation
            ):
                msg = self.cmdchar + msg[nickl + 1 :].lstrip()
        else:
            # Turn 'nick:' prefixes into self.cmdchar prefixes
            if (
                lmsg.startswith(lnick)
                and len(lmsg) > nickl
                and lmsg[nickl] in string.punctuation
            ):
                msg = self.cmdchar + msg[len(self.nickname) + 1 :].lstrip()
        reply = (channel == lnick) and user or channel

        if msg.startswith(self.cmdchar):
            cmnd = msg[len(self.cmdchar) :]
            self._command(user, reply, cmnd)

        # Run privmsg handlers
        self._runhandler("privmsg", user, reply, self.factory.to_unicode(msg))

        # run URL handlers
        urls = pyfiurl.grab(msg)
        if urls:
            for url in urls:
                self._runhandler("url", user, reply, url, self.factory.to_unicode(msg))

    def _runhandler(self, handler, *args, **kwargs):
        """Run a handler for an event"""
        handler = "handle_%s" % handler
        # module commands
        for module, env in self.factory.ns.items():
            myglobals, mylocals = env
            # find all matching command functions
            handlers = [
                (h, ref)
                for h, ref in mylocals.items()
                if h == handler and isinstance(ref, FunctionType)
            ]

            for hname, func in handlers:
                # defer each handler to a separate thread, assign callbacks to see when they end
                d = threads.deferToThread(func, self, *args, **kwargs)
                d.addCallback(self.printResult, "handler %s completed" % hname)
                d.addErrback(self.printError, "handler %s error" % hname)

    def _runEvents(self, eventname, *args, **kwargs):
        """Run funtions on events named by eventname parameter"""
        eventname = "event_%s" % eventname
        for module, env in self.factory.ns.items():
            myglobals, mylocals = env

            # find all matching events
            events = [
                (h, ref)
                for h, ref in mylocals.items()
                if h == eventname and isinstance(ref, FunctionType)
            ]

            for ename, func in events:
                # defer each handler to a separate thread, assign callbacks to see when they end
                d = threads.deferToThread(func, self, *args, **kwargs)
                d.addCallback(
                    self.printResult, "%s %s event completed" % (module, ename)
                )
                d.addErrback(self.printError, "%s %s event error" % (module, ename))

    def _command(self, user, channel, cmnd):
        """Handles bot commands.

        This function calls the appropriate method for the given command.

        The command methods are formatted as "command_<commandname>"
        """
        # Split arguments from the command part
        try:
            cmnd, args = cmnd.split(" ", 1)
        except ValueError:
            args = ""

        # core commands
        method = getattr(self, "command_%s" % cmnd, None)
        if method is not None:
            log.info(
                "internal command %s called by %s (%s) on %s"
                % (cmnd, user, self.factory.isAdmin(user), channel)
            )
            method(user, channel, args)
            return

        # module commands
        for module, env in self.factory.ns.items():
            myglobals, mylocals = env
            # find all matching command functions
            commands = [
                (c, ref) for c, ref in mylocals.items() if c == "command_%s" % cmnd
            ]
            commands += [
                (c, ref) for c, ref in mylocals.items() if c == "admin_%s" % cmnd
            ]

            for cname, command in commands:
                if not self.factory.isAdmin(user) and cname.startswith("admin"):
                    continue
                i = inspect.getcallargs(command, self, user, channel, cmnd)
                if "silent" not in i:
                    log.info(
                        "module command %s called by %s (%s) on %s"
                        % (cname, user, self.factory.isAdmin(user), channel)
                    )
                # Defer commands to threads
                d = threads.deferToThread(
                    command, self, user, channel, self.factory.to_unicode(args.strip())
                )
                d.addCallback(self.printResult, "command %s completed" % cname)
                d.addErrback(self.printError, "command %s error" % cname)

    # Overrides for twisted.words.irc core commands #
    def say(
        self, channel: str, message: str, length: Optional[int] = None
    ) -> Tuple[str, str, str]:
        """Override default say to make replying to private messages easier"""

        # Ensure channel is a string
        # (for cases where channel is specified in code instead of "answering")
        channel = self.factory.to_unicode(channel)
        # Ensure all outgoing messages are strings
        message = self.factory.to_unicode(message)

        # Change nick!user@host -> nick, since all servers don't support full hostmask messaging
        if "!" and "@" in channel:
            channel = self.get_nick(channel)

        # wrap long text into suitable fragments
        msg = self.tw.wrap(message)
        cont = False

        for m in msg:
            if cont:
                m = "..." + m
            self.msg(channel, m, length)
            cont = True

        return ("botcore.say", channel, message)

    def act(self, channel, message, length=None):
        """Use act instead of describe for actions"""
        return super(PyFiBot, self).describe(channel, message)

    def mode(self, chan, set, modes, limit=None, user=None, mask=None):
        chan = self.factory.to_unicode(chan)
        _set = self.factory.to_unicode(set)
        modes = self.factory.to_unicode(modes)
        return super(PyFiBot, self).mode(chan, _set, modes, limit, user, mask)

    def kick(self, channel, user, reason=None):
        reason = self.factory.to_unicode(reason)
        return super(PyFiBot, self).kick(channel, user, reason)

    def join(self, channel, key=None):
        channel = self.factory.to_unicode(channel)
        return super(PyFiBot, self).join(channel, key)

    def leave(self, channel, key=None):
        channel = self.factory.to_unicode(channel)
        return super(PyFiBot, self).leave(channel, key)

    def quit(self, message=""):
        message = self.factory.to_unicode(message)
        return super(PyFiBot, self).quit(message)

    # Overrides for twisted.words.irc internal commands #
    def XXregister(self, nickname, hostname="foo", servername="bar"):
        nickname = self.factory.to_unicode(nickname)
        hostname = self.factory.to_unicode(hostname)
        servername = self.factory.to_unicode(servername)
        return super(PyFiBot, self).register(nickname, hostname, servername)

        # self.sendLine("USER %s %s %s :%s" % (self.username, hostname, servername, self.realname))
        # self.register(nickname, hostname, servername)

    # LOW-LEVEL IRC HANDLERS #

    def irc_JOIN(self, prefix, params):
        """override the twisted version to preserve full userhost info"""
        nick = self.get_nick(prefix)
        channel = params[-1]

        if nick == self.nickname:
            self.joined(channel)
        else:
            self.userJoined(prefix, channel)

        if nick.lower() != self.nickname.lower():
            pass
        elif channel not in self.network.channels:
            self.network.channels.append(channel)
            self.factory.setNetwork(self.network)

    def irc_PART(self, prefix, params):
        """override the twisted version to preserve full userhost info"""

        nick = self.get_nick(prefix)
        channel = params[0]

        if nick == self.nickname:
            self.left(channel)
        else:
            # some clients don't send a part message at all, compensate
            if len(params) == 1:
                params.append("")
            self.userLeft(prefix, channel, params[1])

    def irc_NICK(self, prefix, params):
        """override the twisted version to preserve full userhost info"""
        newnick = params[0]
        self.userRenamed(prefix, newnick)

    def irc_QUIT(self, prefix, params):
        """QUIT-handler.

        Twisted IRCClient doesn't handle this at all.."""

        nick = self.get_nick(prefix)
        if nick == self.nickname:
            self.left(params)
        else:
            self.userLeft(prefix, None, params[0])

    # HANDLERS #

    # ME #
    def joined(self, channel):
        """I joined a channel"""
        self._runhandler("joined", channel)

    def left(self, channel):
        """I left a channel"""
        self._runhandler("left", channel)

    def noticed(self, user, channel, message):
        """I received a notice"""
        self._runhandler("noticed", user, channel, self.factory.to_unicode(message))

    def modeChanged(self, user, channel, set, modes, args):
        """Mode changed on user or channel"""
        self._runhandler("modeChanged", user, channel, set, modes, args)

    def kickedFrom(self, channel, kicker, message):
        """I was kicked from a channel"""
        self._runhandler(
            "kickedFrom", channel, kicker, self.factory.to_unicode(message)
        )

    def nickChanged(self, nick):
        """I changed my nick"""
        self._runhandler("nickChanged", nick)

    # OTHER PEOPLE
    def userJoined(self, user, channel):
        """Someone joined"""
        self._runhandler("userJoined", user, channel)

    def userLeft(self, user, channel, message):
        """Someone left"""
        self._runhandler("userLeft", user, channel, self.factory.to_unicode(message))

    def userKicked(self, kickee, channel, kicker, message):
        """Someone got kicked by someone"""
        self._runhandler(
            "userKicked", kickee, channel, kicker, self.factory.to_unicode(message)
        )

    def action(self, user, channel, data):
        """An action"""
        self._runhandler("action", user, channel, self.factory.to_unicode(data))

    def topicUpdated(self, user, channel, topic):
        """Save topic to maindb when it changes"""
        self._runhandler("topicUpdated", user, channel, self.factory.to_unicode(topic))

    def userRenamed(self, oldnick, newnick):
        """Someone changed their nick"""
        self._runhandler("userRenamed", oldnick, newnick)

    def receivedMOTD(self, motd):
        """MOTD"""
        self._runhandler("receivedMOTD", self.factory.to_unicode(motd))

    # SERVER INFORMATION

    # Network = Quakenet -> do Q auth
    def isupport(self, options):
        log.info(self.network.alias + " SUPPORTS: " + ",".join(options))

    def created(self, when):
        log.info(self.network.alias + " CREATED: " + when)

    def yourHost(self, info):
        log.info(self.network.alias + " YOURHOST: " + info)

    def myInfo(self, servername, version, umodes, cmodes):
        log.info(
            self.network.alias
            + " MYINFO: %s %s %s %s" % (servername, version, umodes, cmodes)
        )

    def luserMe(self, info):
        log.info(self.network.alias + " LUSERME: " + info)



# -*- coding: utf-8 -*-
from nose.tools import eq_
import bot_mock
from pyfibot.modules import module_urltitle

import pytest
from vcr import VCR

my_vcr = VCR(
    path_transformer=VCR.ensure_suffix(".yaml"),
    cassette_library_dir="tests/cassettes/",
    record_mode=pytest.config.getoption("--vcrmode"),
)


@pytest.fixture
def botmock():
    bot = bot_mock.BotMock()
    module_urltitle.init(bot)
    return bot


@my_vcr.use_cassette()
def test_one(botmock):
    msg = "https://en.wikipedia.org/wiki/Hatfield–McCoy_feud"
    assert (
        "#channel",
        "Title: The Hatfield\u2013McCoy feud, also described by journalists as the Hatfield\u2013McCoy war, involved two rural American families of the West Virginia\u2013Kentucky area along the Tug Fork of the Big Sandy River in the years 1863\u20131891.",
    ) == module_urltitle.handle_url(botmock, None, "#channel", msg, msg)


@my_vcr.use_cassette()
def test_two(botmock):
    msg = "http://fi.wikipedia.org/wiki/DTMF"
    assert (
        "#channel",
        "Title: DTMF on puhelinlaitteissa k\xe4ytetty numeroiden \xe4\xe4nitaajuusvalinta.",
    ) == module_urltitle.handle_url(botmock, None, "#channel", msg, msg)


@my_vcr.use_cassette()
def test_three(botmock):
    msg = "http://en.wikipedia.org/wiki/Gender_performativity"
    assert (
        "#channel",
        "Title: The social construction of gender is a theory in feminism and sociology about the manifestation of cultural origins, mechanisms, and corollaries of gender perception and expression in the context of interpersonal and group social interaction.",
    ) == module_urltitle.handle_url(botmock, None, "#channel", msg, msg)


@my_vcr.use_cassette()
def test_four(botmock):
    msg = "http://en.wikipedia.org/wiki/Dynamo_(magician)"
    eq_(
        (
            "#channel",
            "Title: Steven Frayne, better known by his stage name Dynamo, is a British magician born in Bradford, West Yorkshire.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )


@my_vcr.use_cassette()
def test_five(botmock):
    msg = "http://fi.wikipedia.org/wiki/David_Eddings"
    eq_(
        (
            "#channel",
            "Title: David Carroll Eddings oli yhdysvaltalainen kirjailija, joka kirjoitti useita suosittuja fantasiakirjoja.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )


@my_vcr.use_cassette()
def test_six(botmock):
    msg = "http://fi.wikipedia.org/wiki/Birger_Ek"
    module_urltitle.init(botmock)
    eq_(
        (
            "#channel",
            "Title: Rolf Birger Ek oli suomalainen lentäjä ja Mannerheim-ristin ritari.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )


@my_vcr.use_cassette()
def test_seven(botmock):
    msg = "http://en.wikipedia.org/wiki/Ramon_Llull"
    eq_(
        (
            "#channel",
            "Title: Ramon Llull was a mathematician, polymath, philosopher, logician, writer and mystic from the Kingdom of Majorca.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )


@my_vcr.use_cassette()
def test_eight(botmock):
    msg = "https://en.wikipedia.org/wiki/Byzantine_art#Periods"
    eq_(
        (
            "#channel",
            "Title: Byzantine art comprises the body of Christian Greek artistic products of the Eastern Roman Empire, as well as the nations and states that inherited culturally from the empire.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )


@my_vcr.use_cassette()
def test_nine(botmock):
    msg = "http://fi.wikipedia.org/wiki/Kimi_Räikkönen"
    eq_(
        (
            "#channel",
            "Title: Kimi-Matias Räikkönen on suomalainen autourheilija ja Formula 1 -sarjan maailmanmestari.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )


@my_vcr.use_cassette()
def test_ten(botmock):
    msg = "http://en.wikipedia.org/wiki/802.11ac"
    eq_(
        (
            "#channel",
            "Title: IEEE 802.11ac-2013 or 802.11ac is a wireless networking standard in the 802.11 set of protocols, providing high-throughput wireless local area networks on the 5\xa0GHz band.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )


@my_vcr.use_cassette()
def test_eleven(botmock):
    msg = "http://en.wikipedia.org/wiki/Edison_Arantes_do_Nascimento"
    eq_(
        (
            "#channel",
            "Title: Edson Arantes do Nascimento, known as Pel\xe9, is a Brazilian former professional footballer who played as a forward.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )


@my_vcr.use_cassette()
def test_twelve(botmock):
    msg = "http://en.wikipedia.org/wiki/Mr._Bean"
    eq_(
        (
            "#channel",
            "Title: Mr. Bean is a British sitcom created by Rowan Atkinson and Richard Curtis, produced by Tiger Aspect and starring Atkinson as the title character.",
        ),
        module_urltitle.handle_url(botmock, None, "#channel", msg, msg),
    )



# -*- coding: utf-8 -*-
import pytest
import sys
import os
import socket
from unittest.mock import Mock, patch

# Add the project root to Python path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from pyfibot.modules.module_webchat import webchat_getorigin
from pyfibot.modules.module_autoop import check_hostmask

# Try to import wolfram alpha functions
try:
    from pyfibot.modules.module_wolfram_alpha import clean_question, clean_answer
    WOLFRAM_AVAILABLE = True
except ImportError:
    WOLFRAM_AVAILABLE = False


class TestWebchatOrigin:
    """Test webchat hex IP origin parsing"""

    def test_valid_hex_ip(self):
        """Test parsing valid hex IP addresses"""
        # Test case: C0A80001 -> 192.168.0.1
        result = webchat_getorigin("C0A80001")
        assert result is not None
        assert "192.168.0.1" in result

    def test_invalid_hex_length(self):
        """Test invalid hex IP length"""
        # Too short
        assert webchat_getorigin("C0A800") is None
        # Too long
        assert webchat_getorigin("C0A800011") is None
        # Empty
        assert webchat_getorigin("") is None

    def test_invalid_hex_characters(self):
        """Test invalid hex characters"""
        # Contains non-hex characters
        assert webchat_getorigin("G0A80001") is None
        assert webchat_getorigin("C0A8000Z") is None
        assert webchat_getorigin("HELLO123") is None

    def test_valid_common_ips(self):
        """Test common IP addresses in hex format"""
        test_cases = [
            ("7F000001", "127.0.0.1"),  # localhost
            ("C0A80101", "192.168.1.1"),  # common router IP
            ("08080808", "8.8.8.8"),     # Google DNS
        ]
        
        for hex_ip, expected_ip in test_cases:
            result = webchat_getorigin(hex_ip)
            assert result is not None
            assert expected_ip in result

    def test_zero_ip(self):
        """Test all-zero IP address"""
        result = webchat_getorigin("00000000")
        assert result is not None
        assert "0.0.0.0" in result

    def test_max_ip(self):
        """Test maximum IP address (255.255.255.255)"""
        result = webchat_getorigin("FFFFFFFF")
        assert result is not None
        assert "255.255.255.255" in result

    def test_lowercase_hex(self):
        """Test lowercase hex input"""
        result = webchat_getorigin("c0a80001")
        assert result is not None
        assert "192.168.0.1" in result

    def test_mixed_case_hex(self):
        """Test mixed case hex input"""
        result = webchat_getorigin("C0a80001")
        assert result is not None
        assert "192.168.0.1" in result

    @patch('socket.getfqdn')
    def test_hostname_resolution(self, mock_getfqdn):
        """Test hostname resolution"""
        mock_getfqdn.return_value = "example.com"
        
        result = webchat_getorigin("C0A80001")
        assert result is not None
        assert "192.168.0.1 -> example.com" in result

    @patch('socket.getfqdn')
    def test_no_hostname_resolution(self, mock_getfqdn):
        """Test when hostname resolution returns same as IP"""
        mock_getfqdn.return_value = "192.168.0.1"  # Same as IP
        
        result = webchat_getorigin("C0A80001")
        assert result is not None
        assert "192.168.0.1" in result
        assert " -> " not in result  # Should not show arrow

    @patch('pyfibot.modules.module_webchat.gi4')
    @patch('socket.getfqdn')
    def test_geoip_integration(self, mock_getfqdn, mock_gi4):
        """Test GeoIP integration when available"""
        mock_getfqdn.return_value = "192.168.0.1"
        mock_gi4.country_name_by_addr.return_value = "Finland"
        
        result = webchat_getorigin("C0A80001")
        assert result is not None
        assert "192.168.0.1" in result
        assert "(Finland)" in result

    @patch('pyfibot.modules.module_webchat.gi4')
    @patch('socket.getfqdn')
    def test_geoip_failure(self, mock_getfqdn, mock_gi4):
        """Test GeoIP failure handling"""
        mock_getfqdn.return_value = "192.168.0.1"
        mock_gi4.country_name_by_addr.side_effect = socket.gaierror()
        
        result = webchat_getorigin("C0A80001")
        assert result is not None
        assert "192.168.0.1" in result
        # Should not fail even if GeoIP fails

    def test_edge_case_ips(self):
        """Test edge case IP addresses"""
        edge_cases = [
            "00000001",  # 0.0.0.1
            "FF000001",  # 255.0.0.1
            "00FF0001",  # 0.255.0.1
            "0000FF01",  # 0.0.255.1
            "000000FF",  # 0.0.0.255
        ]
        
        for hex_ip in edge_cases:
            result = webchat_getorigin(hex_ip)
            assert result is not None
            assert "." in result  # Should contain IP format


class TestHostmaskValidation:
    """Test hostmask validation"""

    def test_valid_hostmasks(self):
        """Test valid hostmask formats"""
        valid_hostmasks = [
            "nick!user@host.com",
            "testnick!testuser@example.org",
            "bot!~botuser@bot.domain.net",
            "user123!user@192.168.1.1",
            "nick!ident@gateway/web/freenode",
            "someone!~someone@unaffiliated/someone",
        ]
        
        for hostmask in valid_hostmasks:
            assert check_hostmask(hostmask) is True

    def test_invalid_hostmasks(self):
        """Test invalid hostmask formats"""
        invalid_hostmasks = [
            "nick",  # Missing ! and @
            "nick!user",  # Missing @
            "nick@host.com",  # Missing !
            "!user@host.com",  # Missing nick
            "nick!@host.com",  # Missing user
            "nick!user@",  # Missing host
            "",  # Empty string
            "nick user@host.com",  # Space instead of !
            "nick!user host.com",  # Space instead of @
        ]
        
        for hostmask in invalid_hostmasks:
            assert check_hostmask(hostmask) is False

    def test_edge_case_hostmasks(self):
        """Test edge case hostmasks"""
        edge_cases = [
            "a!b@c",  # Minimal valid hostmask
            "nick!!user@host",  # Double !
            "nick!user@@host",  # Double @
            "nick!user@host@extra",  # Multiple @
            "nick!user!extra@host",  # Multiple !
        ]
        
        for hostmask in edge_cases:
            result = check_hostmask(hostmask)
            # These should still match the regex pattern
            assert isinstance(result, bool)

    def test_unicode_hostmasks(self):
        """Test hostmasks with unicode characters"""
        unicode_hostmasks = [
            "niçk!üser@höst.com",
            "тест!пользователь@хост.рф",
            "用户!身份@主机.中国",
        ]
        
        for hostmask in unicode_hostmasks:
            result = check_hostmask(hostmask)
            # Should handle unicode characters
            assert isinstance(result, bool)

    def test_very_long_hostmasks(self):
        """Test very long hostmasks"""
        long_nick = "a" * 100
        long_user = "b" * 100
        long_host = "c" * 100 + ".com"
        long_hostmask = f"{long_nick}!{long_user}@{long_host}"
        
        assert check_hostmask(long_hostmask) is True

    def test_special_characters_in_hostmask(self):
        """Test hostmasks with special characters"""
        special_cases = [
            "nick[bot]!user@host.com",
            "nick-test!user_test@host-name.com",
            "nick.bot!~user@host.example.net",
            "nick123!user456@192.168.1.100",
        ]
        
        for hostmask in special_cases:
            assert check_hostmask(hostmask) is True


@pytest.mark.skipif(not WOLFRAM_AVAILABLE, reason="Wolfram Alpha module not available")
class TestWolframStringCleaning:
    """Test Wolfram Alpha string cleaning functions"""

    def test_clean_answer_basic(self):
        """Test basic string cleaning"""
        assert clean_answer("simple text") == "simple text"
        assert clean_answer("") == ""
        assert clean_answer(None) is None

    def test_clean_answer_pipe_replacement(self):
        """Test pipe character replacement"""
        input_str = "value1 | value2 | value3"
        result = clean_answer(input_str)
        assert result == "value1: value2: value3"

    def test_clean_answer_newline_replacement(self):
        """Test newline replacement"""
        input_str = "line1\nline2\nline3"
        result = clean_answer(input_str)
        assert result == "line1 | line2 | line3"

    def test_clean_answer_tilde_replacement(self):
        """Test tilde replacement with approximation"""
        input_str = "value ~~ 3.14"
        result = clean_answer(input_str)
        assert "≈" in result
        assert "3.14" in result

    def test_clean_answer_multiple_spaces(self):
        """Test multiple space reduction"""
        input_str = "word1    word2     word3"
        result = clean_answer(input_str)
        assert result == "word1 word2 word3"

    def test_clean_answer_whitespace_trimming(self):
        """Test whitespace trimming"""
        input_str = "  text with spaces  "
        result = clean_answer(input_str)
        assert result == "text with spaces"

    def test_clean_answer_currency_symbols(self):
        """Test currency symbol replacement"""
        # Bitcoin symbol
        input_str = "Price: \\:0e3f 1000"
        result = clean_answer(input_str)
        assert "฿" in result
        assert "1000" in result

        # Yen symbol
        input_str = "Price: \\:ffe5 500"
        result = clean_answer(input_str)
        assert "￥" in result
        assert "500" in result

    def test_clean_answer_complex_example(self):
        """Test complex cleaning with multiple operations"""
        input_str = "Result | answer\nValue ~~ 42   | Currency: \\:0e3f 100"
        result = clean_answer(input_str)
        
        # Should apply all transformations
        assert "Result: answer" in result
        assert "Value ≈ 42" in result
        assert "Currency: ฿ 100" in result
        assert "\n" not in result

    def test_clean_question_delegates_to_clean_answer(self):
        """Test that clean_question uses clean_answer"""
        test_string = "question | with\nnewlines"
        question_result = clean_question(test_string)
        answer_result = clean_answer(test_string)
        assert question_result == answer_result

    def test_clean_answer_edge_cases(self):
        """Test edge cases for string cleaning"""
        # Only special characters
        assert clean_answer("~~~") == "≈≈≈"
        assert clean_answer("|||") == ":::"
        assert clean_answer("\n\n\n") == " |  | "

        # Mixed operations
        mixed = "~~~ | \n   multiple   spaces"
        result = clean_answer(mixed)
        assert "≈≈≈" in result
        assert ":" in result
        assert "|" in result
        # Should not have multiple consecutive spaces
        assert "  " not in result

    def test_clean_answer_mathematical_expressions(self):
        """Test cleaning of mathematical expressions"""
        math_expr = "x ~~ 3.14159 | y = 2.71828\nz ~~ 1.41421"
        result = clean_answer(math_expr)
        
        assert "x ≈ 3.14159" in result
        assert "y = 2.71828" in result
        assert "z ≈ 1.41421" in result
        assert " | " in result  # Newline converted to pipe
        assert ":" in result  # Pipe converted to colon

    def test_clean_answer_real_wolfram_output(self):
        """Test with realistic Wolfram Alpha output"""
        wolfram_output = """Result | 42
        Input interpretation | number
        
        Result ~~ 4.2 × 10^1"""
        
        result = clean_answer(wolfram_output)
        
        assert "Result: 42" in result
        assert "Input interpretation: number" in result
        assert "4.2 × 10^1" in result
        assert "≈" in result
        assert "\n" not in result


class TestModuleHelperEdgeCases:
    """Test edge cases and error conditions for module helpers"""

    def test_webchat_getorigin_memory_efficiency(self):
        """Test that webchat_getorigin doesn't use excessive memory"""
        # Test with many invalid inputs to ensure no memory leaks
        invalid_inputs = [
            "G" * 8,  # Invalid hex
            "1" * 7,  # Wrong length
            "ZZZZZZZZ",  # Invalid hex
            "",  # Empty
            None,  # This might cause an error, but shouldn't crash
        ]
        
        for invalid_input in invalid_inputs:
            try:
                result = webchat_getorigin(invalid_input)
                assert result is None
            except (TypeError, AttributeError):
                # None input might cause these, which is acceptable
                pass

    def test_hostmask_regex_performance(self):
        """Test hostmask validation with various edge cases"""
        # Very long strings
        very_long = "a" * 10000
        result = check_hostmask(very_long)
        assert isinstance(result, bool)
        
        # String with many special characters
        special_string = "!@#$%^&*()!user@host.com"
        result = check_hostmask(special_string)
        assert isinstance(result, bool)

    @pytest.mark.skipif(not WOLFRAM_AVAILABLE, reason="Wolfram Alpha module not available")
    def test_clean_answer_performance(self):
        """Test string cleaning with large inputs"""
        # Large string with many replacements
        large_string = ("text | text\n" * 1000) + ("value ~~ number " * 1000)
        result = clean_answer(large_string)
        
        assert isinstance(result, str)
        assert len(result) > 0
        assert ":" in result  # Pipes should be replaced
        assert "≈" in result  # Tildes should be replaced

    def test_integration_realistic_usage(self):
        """Test functions with realistic usage patterns"""
        # Simulate realistic webchat usage
        common_hex_ips = [
            "C0A80001",  # 192.168.0.1
            "0A000001",  # 10.0.0.1
            "AC100001",  # 172.16.0.1
        ]
        
        for hex_ip in common_hex_ips:
            result = webchat_getorigin(hex_ip)
            assert result is not None
            assert "." in result  # Should contain IP
        
        # Simulate realistic hostmask validation
        common_hostmasks = [
            "user!~webchat@gateway/web/freenode",
            "nick!user@unaffiliated/nick",
            "bot!bot@example.com",
        ]
        
        for hostmask in common_hostmasks:
            assert check_hostmask(hostmask) is True


# -*- coding: utf-8 -*-
"""Search Youtube videos
Returns the first Youtube video found with the given keywords using Youtube API.
"""

from __future__ import unicode_literals, print_function, division


def command_youtube(bot, user, channel, args):
    """<message> Returns the first Youtube video found with the given keywords. Usage: .youtube sultans of swing"""
    search = args

    api_key = config.get("youtube_apikey", None)
    if api_key is None:
        log.warning("Set API key in configuration to activate YouTube API titles")
        return None  # Can't do anything without key

    # replacing spaces with +
    search = search.replace(" ", "+")

    if len(search) > 0:

        api_url = "https://www.googleapis.com/youtube/v3/search?"
        params = {"q": search, "part": "snippet", "type": "video", "key": api_key}

        r = bot.get_url(api_url, params=params)
        j = r.json()

        if j["items"][0]["id"]["videoId"]:
            return bot.say(
                channel,
                j["items"][0]["snippet"]["title"]
                + " - https://www.youtube.com/watch?v="
                + j["items"][0]["id"]["videoId"],
            )
        else:
            return bot.say(channel, "No results")
    else:
        return bot.say(channel, "No results")



"""
Parse spotify URLs

A modified version of the original pyfibot spotify module

Spotify closed down the API and it needs authentication. This works around it by making a request to
an endpoint with a smartphone user-agent. It seems to yield more html-content from which to parse from

Use freely and at own risk. I am not viable for anything.

done in Jan 2018 by T-101 / Darklite ^ Primitive

"""

from __future__ import unicode_literals, print_function, division
import re
import json
import logging

from bs4 import BeautifulSoup

log = logging.getLogger("spotify")


def handle_privmsg(bot, user, channel, args):
    class SpotifyData(object):
        def __init__(self):
            self._data = None
            self.output = None

        @staticmethod
        def _parse_script_elements(scripts):
            """

            Selects the desired <script> element containing the desired json data

            """
            for script in scripts:
                if re.search(r"Spotify.Entity", script.string):
                    return re.sub(r"[\n\t]", "", script.string)

        @staticmethod
        def _ms_human_readable(ms):
            """

            Converts milliseconds into a tuple of (minutes, seconds)

            """
            minutes = int(ms / 60000)
            seconds = int((ms - minutes * 60000) / 1000)
            return minutes, seconds

        @property
        def data(self):
            return self._data

        @data.setter
        def data(self, value):
            """

            Takes a list of <script> elements, selects the desired one and morphs contents to a neat json dict

            """
            soup = BeautifulSoup(value, "html.parser")
            script = self._parse_script_elements(soup.find_all("script"))
            if script:
                element = re.search(r"Spotify.Entity = ({.*});", script).group(1)
                self._data = json.loads(element)

        @data.deleter
        def data(self):
            self._data = None
            self.data = None

        def parse(self, data, data_type):
            """

            Parse data, and set the .output parameter accordingly

            """

            self.data = data
            if not self.data:
                return

            if data_type == "track":
                artist = ", ".join(
                    [artist.get("name") for artist in self.data.get("artists")]
                )
                name = self.data.get("name")
                duration = self._ms_human_readable(int(self.data.get("duration_ms")))

                self.output = "\002[Spotify]\002 %s - %s [%sm%ss]" % (
                    artist,
                    name,
                    duration[0],
                    duration[1],
                )
                if self.data.get("album") and self.data.get("album").get("name"):
                    self.output += ' (from "%s")' % self.data.get("album").get("name")

            elif data_type == "album":
                artist = ", ".join(
                    [artist.get("name") for artist in self.data.get("artists")]
                )
                name = self.data.get("name")
                tracks = self.data.get("tracks").get("total")
                released = 0
                if re.match(r"\d{4}", self.data.get("release_date")):
                    released = re.match(r"\d{4}", self.data.get("release_date")).group()

                self.output = "\002[Spotify Album]\002 %s (%s, %s tracks) by %s" % (
                    name,
                    released,
                    tracks,
                    artist,
                )

            elif data_type == "artist":
                self.output = "\002[Spotify Artist]\002 %s" % self.data.get("name")

            elif data_type == "playlist":
                name = self.data.get("name")
                owner = self.data.get("owner").get("id")
                followers = self.data.get("followers").get("total")
                tracks = self.data.get("tracks").get("total")

                self.output = (
                    "\002[Spotify Playlist]\002 %s by %s (%s tracks, %s followers)"
                    % (name, owner, tracks, followers)
                )

            return self

    m = re.match(
        r".*(https?:\/\/open.spotify.com\/|spotify:)(?P<item>album|artist|track|user[:\/]\S+[:\/]playlist)[:\/](?P<id>[a-zA-Z0-9]+)\/?.*",
        args,
    )
    if not m:
        return None

    spotify_id = m.group("id")
    item = m.group("item").replace(":", "/").split("/")

    if item[-1] not in ("track", "album", "artist", "playlist"):
        bot.log("Given endpoint not yet supported")
        return

    r = re.search(r"(https?:\/\/\S+)", args)
    if r:
        url = r.group()
    elif item[-1] == "playlist":
        url = "https://open.spotify.com/user/%s/playlist/%s" % (item[1], spotify_id)
    else:
        url = "https://open.spotify.com/%s/%s" % (item[-1], spotify_id)

    user_agent = {
        "User-Agent": "Mozilla/5.0 (Linux; Android 6.0.1; SM-G900R4 Build/MMB29M; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/55.0.2883.91 Mobile Safari/537.36"
    }
    r = bot.get_url(url=url, headers=user_agent)
    if r.status_code != 200:
        bot.log("module_spotify_noapi.py HTTP status code %s" % r.status_code)
        # bot.log("HTTP response %s" % r.content)
        return

    spotify_data = SpotifyData().parse(r.content, item[-1])

    if spotify_data.output:
        bot.say(channel, spotify_data.output)



"""
Display webchat users' actual hostnames on join and when requested
"""

from __future__ import unicode_literals, print_function, division
import socket

try:
    from modules.module_geoip import gi4

    geoip_available = True
except ImportError:
    geoip_available = False


def handle_userJoined(bot, user, channel):
    nick = bot.getNick(user)
    userhost = user.split("!")[1]
    username, host = userhost.split("@")
    username = username.replace("~", "").replace("-", "")
    # known webchat hosts
    if host in [
        "webchat.xs4all.nl",
        "wwwirc.kapsi.fi",
        "webchat.mibbit.com",
        "gateway/web/freenode",
        "webchat.ircnet.net",
    ] or host.endswith(".kiwiirc.com"):
        origin = webchat_getorigin(username)
        if origin:
            return bot.say(channel, "%s is using webchat from %s" % (nick, origin))


def command_webchat(bot, user, channel, args):
    """Parse a webchat hex ip to a domain"""
    origin = webchat_getorigin(args)
    if origin:
        return bot.say(channel, "webchat from %s" % origin)
    else:
        return bot.say(
            channel, "%s: %s is not a valid webchat hex ip" % (bot.getNick(user), args)
        )


def webchat_getorigin(hexip):
    """Parse webchat hex-format ip to decimal ip and hostname if it exists"""
    if len(hexip) != 8:
        return

    ip = []
    for i in range(2, len(hexip) + 2, 2):
        try:
            dec = int(hexip[i - 2 : i], 16)
        except ValueError:
            return
        ip.append(str(dec))

    if ip:
        addr = ".".join(ip)
        hostname = socket.getfqdn(addr)
        if hostname != addr:
            origin = "%s -> %s" % (addr, hostname)
        else:
            origin = addr

        try:
            country = gi4.country_name_by_addr(addr)
            origin += " (%s)" % country
        except socket.gaierror:
            pass

    return origin



from __future__ import unicode_literals, print_function, division
import pygeoip
import os.path
import sys
import socket

try:
    from modules.module_usertrack import get_table

    user_track_available = True
except ImportError:
    user_track_available = False


# http://dev.maxmind.com/geoip/legacy/geolite/
DATAFILE = os.path.join(sys.path[0], "GeoIP.dat")

# STANDARD = reload from disk
# MEMORY_CACHE = load to memory
# MMAP_CACHE = memory using mmap
gi4 = pygeoip.GeoIP(DATAFILE, pygeoip.MEMORY_CACHE)


def command_geoip(bot, user, channel, args):
    """Determine the user's country based on host or nick, if module_usertrack is used."""
    if not args:
        return bot.say(channel, "usage: .geoip HOST/NICK")

    host = args
    nick = None

    if user_track_available:
        table = get_table(bot, channel)
        user = table.find_one(nick=args)
        if user:
            nick = user["nick"]
            host = user["host"]

    try:
        country = gi4.country_name_by_name(host)
    except socket.gaierror:
        country = None

    if country:
        if nick:
            return bot.say(channel, "%s (%s) is in %s" % (nick, host, country))
        return bot.say(channel, "%s is in %s" % (host, country))
    if nick:
        return bot.say(channel, "Host not found for %s (%s)" % (nick, host))
    return bot.say(channel, "Host not found for %s" % host)



"""
Parse spotify URLs
"""

from __future__ import unicode_literals, print_function, division
import re
import logging
import base64
import requests

log = logging.getLogger("spotify")


def get_token(client_id, client_secret):
    token = client_id + ":" + client_secret
    encoded_token = base64.b64encode(token)

    payload = {"grant_type": "client_credentials"}
    headers = {"Authorization": "Basic " + encoded_token}
    auth_url = "https://accounts.spotify.com/api/token"
    r = requests.post(auth_url, payload, headers=headers)

    return r.json()["access_token"]


def _handle_privmsg(bot, user, channel, args):
    """Grab Spotify URLs from the messages and handle them"""
    config = bot.config.get("module_spotify", {})

    m = re.match(
        r".*(https?:\/\/open.spotify.com\/|spotify:)(?P<item>album|artist|track|user[:\/]\S+[:\/]playlist)[:\/](?P<id>[a-zA-Z0-9]+)\/?.*",
        args,
    )
    if not m:
        return None

    spotify_id = m.group("id")
    item = m.group("item").replace(":", "/").split("/")
    item[0] += "s"
    if item[0] == "users":
        # All playlists seem to return 401 at the time, even the public ones
        return None

    bearer_token = get_token(config.get("client_id"), config.get("client_secret"))

    headers = {"Authorization": "Bearer " + bearer_token}
    apiurl = "https://api.spotify.com/v1/%s/%s" % ("/".join(item), spotify_id)
    r = bot.get_url(apiurl, headers=headers)

    if r.status_code != 200:
        if r.status_code not in [401, 403]:
            log.warning(
                "Spotify API returned %s while trying to fetch %s" % r.status_code,
                apiurl,
            )
        print(r.json())
        return bot.say(channel, "Error: %s" % r.json()["error"]["message"])

    data = r.json()

    title = "[Spotify] "
    if item[0] in ["albums", "tracks"]:
        artists = []
        for artist in data["artists"]:
            artists.append(artist["name"])
        title += ", ".join(artists)

    if item[0] == "albums":
        title += " - %s (%s)" % (data["name"], data["release_date"])

    if item[0] == "artists":
        title += data["name"]
        genres_n = len(data["genres"])
        if genres_n > 0:
            genitive = "s" if genres_n > 1 else ""
            genres = data["genres"][0:4]
            more = " +%s more" % genres_n - 5 if genres_n > 4 else ""

            title += " (Genre%s: %s%s)" % (genitive, ", ".join(genres), more)

    if item[0] == "tracks":
        title += " - %s - %s" % (data["album"]["name"], data["name"])

    return bot.say(channel, title)



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals, print_function, division
from datetime import datetime, timedelta, timezone
from math import isnan
from typing import Optional
import logging

import lxml.etree as etree
import requests_cache
from tenacity import retry, stop_after_attempt, wait_exponential
from pydantic import BaseModel, Field, field_validator

global default_place
default_place = "Helsinki"

log = logging.getLogger("fmi")

# Initialize cached session for API requests (15 minute cache)
cached_session = requests_cache.CachedSession(
    "fmi_cache", expire_after=900, backend="memory"  # 15 minutes
)


class WeatherData(BaseModel):
    """Pydantic model for weather data validation"""

    place: str = Field(..., description="Location name")
    temperature: Optional[float] = Field(None, description="Temperature in Celsius")
    feels_like: Optional[float] = Field(None, description="Feels like temperature")
    wind_speed: Optional[float] = Field(None, description="Wind speed in m/s")
    humidity: Optional[int] = Field(None, description="Relative humidity percentage")
    cloudiness: Optional[int] = Field(None, description="Cloudiness 0-8")
    weather_description: Optional[str] = Field(None, description="Weather description")

    @field_validator("humidity", "cloudiness", mode="before")
    @classmethod
    def validate_percentages(cls, v):
        if v is not None and (v < 0 or v > 100):
            return None
        return v


# Time format for the API
TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
# t2m = temperature
# ws_10min = wind speed (10 min avg)
# rh == relative humidity
# n_man == cloudiness
# wawa == weather description number
PARAMETERS = ["t2m", "ws_10min", "rh", "n_man", "wawa"]

# http://ilmatieteenlaitos.fi/avoin-data-saahavainnot
WAWA = {
    10: "utua",
    20: "sumua",
    21: "sadetta",
    22: "tihkusadetta",
    23: "vesisadetta",
    24: "lumisadetta",
    25: "jäätävää tihkua",
    30: "sumua",
    31: "sumua",
    32: "sumua",
    33: "sumua",
    34: "sumua",
    40: "sadetta",
    41: "heikkoa tai kohtalaista sadetta",
    42: "kovaa sadetta",
    50: "tihkusadetta",
    51: "heikkoa tihkusadetta",
    52: "kohtalaista tihkusadetta",
    53: "kovaa tihkusadetta",
    54: "jäätävää heikkoa tihkusadetta",
    55: "jäätävää kohtalaista tihkusadetta",
    56: "jäätävää kovaa tihkusadetta",
    60: "vesisadetta",
    61: "heikkoa vesisadetta",
    62: "kohtalaista vesisadetta",
    63: "kovaa vesisadetta",
    64: "jäätävää heikkoa vesisadetta",
    65: "jäätävää kohtalaista vesisadetta",
    66: "jäätävää kovaa vesisadetta",
    67: "räntää",
    68: "räntää",
    70: "lumisadetta",
    71: "heikkoa lumisadetta",
    72: "kohtalaista lumisadetta",
    73: "tiheää lumisadetta",
    74: "heikkoa jääjyväsadetta",
    75: "kohtalaista jääjyväsadetta",
    76: "kovaa jääjyväsadetta",
    77: "lumijyväsiä",
    78: "jääkiteitä",
    80: "sadekuuroja",
    81: "heikkoja sadekuuroja",
    82: "kohtalaisia sadekuuroja",
    84: "heikkoja lumikuuroja",
    85: "kohtalaisia lumikuuroja",
    86: "kovia lumikuuroja",
    87: "raekuuroja",
}


@retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10))
def _fetch_weather_data(place: str) -> Optional[WeatherData]:
    """Fetch weather data from FMI API with retry logic and caching"""
    starttime = (datetime.now(timezone.utc) - timedelta(minutes=10)).strftime(
        TIME_FORMAT
    ) + "Z"
    params = {
        "request": "getFeature",
        "storedquery_id": "fmi::observations::weather::timevaluepair",
        "parameters": ",".join(PARAMETERS),
        "crs": "EPSG::3067",
        "place": place,
        "maxlocations": 1,
        "starttime": starttime,
    }

    try:
        response = cached_session.get(
            "http://opendata.fmi.fi/wfs", params=params, timeout=10
        )
        response.raise_for_status()

        # Parse XML with lxml for better performance
        root = etree.fromstring(response.content)

        # Define namespaces
        namespaces = {
            "gml": "http://www.opengis.net/gml/3.2",
            "wml2": "http://www.opengis.net/waterml/2.0",
        }

        # Get FMI name for more accurate location
        name_element = root.find(".//gml:name", namespaces)
        if name_element is None:
            log.warning(f"Location not found: {place}")
            return None

        location_name = name_element.text

        # Extract measurement values
        values = {}
        for mts in root.findall(".//wml2:MeasurementTimeseries", namespaces):
            gml_id = mts.get("{http://www.opengis.net/gml/3.2}id")
            if gml_id:
                target = gml_id.split("-")[-1]
                value_elements = mts.findall(".//wml2:value", namespaces)
                if value_elements:
                    try:
                        value = float(value_elements[-1].text)
                        if not isnan(value):
                            values[target] = value
                    except (ValueError, TypeError):
                        continue

        # Calculate feels like temperature if we have both temp and wind
        feels_like = None
        if "t2m" in values and "ws_10min" in values:
            feels_like = (
                13.12
                + 0.6215 * values["t2m"]
                - 13.956 * (values["ws_10min"] ** 0.16)
                + 0.4867 * values["t2m"] * (values["ws_10min"] ** 0.16)
            )

        # Get weather description
        weather_desc = None
        if "wawa" in values and int(values["wawa"]) in WAWA:
            weather_desc = WAWA[int(values["wawa"])]

        # Create and validate weather data
        weather_data = WeatherData(
            place=location_name,
            temperature=values.get("t2m"),
            feels_like=feels_like,
            wind_speed=values.get("ws_10min"),
            humidity=int(values["rh"]) if "rh" in values else None,
            cloudiness=int(values["n_man"]) if "n_man" in values else None,
            weather_description=weather_desc,
        )

        return weather_data

    except Exception as e:
        log.error(f"Error fetching weather data for {place}: {e}")
        raise


def init(bot):
    global default_place
    config = bot.config.get("module_fmi", {})
    default_place = config.get("default_place", default_place)


def command_saa(bot, user, channel, args):
    """Command to fetch data from FMI with enhanced error handling and caching"""
    place = args if args else default_place

    try:
        weather_data = _fetch_weather_data(place)
        if not weather_data:
            return bot.say(channel, "Paikkaa ei löytynyt.")

        # Build text from weather data
        text_parts = []

        if weather_data.temperature is not None:
            text_parts.append(f"lämpötila: {weather_data.temperature:.1f}°C")

        if weather_data.feels_like is not None:
            text_parts.append(f"tuntuu kuin: {weather_data.feels_like:.1f}°C")

        if weather_data.wind_speed is not None:
            text_parts.append(f"tuulen nopeus: {round(weather_data.wind_speed)} m/s")

        if weather_data.humidity is not None:
            text_parts.append(f"ilman kosteus: {weather_data.humidity}%%")

        if weather_data.cloudiness is not None:
            text_parts.append(f"pilvisyys: {weather_data.cloudiness}/8")

        if weather_data.weather_description:
            text_parts.append(weather_data.weather_description)

        if not text_parts:
            return bot.say(channel, f"{weather_data.place}: ei säätietoja saatavilla")

        response = f"{weather_data.place}: {', '.join(text_parts)}"
        return bot.say(channel, response)

    except Exception as e:
        log.error(f"Weather command failed for {place}: {e}")
        return bot.say(channel, f"Säätietojen haku epäonnistui: {place}")


def command_keli(bot, user, channel, args):
    """Alias for command "saa" """
    return command_saa(bot, user, channel, args)



from __future__ import unicode_literals, print_function, division
import subprocess
import sys

import logging
from git import Repo, GitCommandError
from tenacity import retry, stop_after_attempt, wait_exponential

log = logging.getLogger("update")


@retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10))
def _git_pull(repo_path):
    """Pull from git with retry logic"""
    repo = Repo(repo_path)
    origin = repo.remotes.origin
    return origin.pull()


@retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10))
def _install_dependencies(cwd):
    """Install/upgrade dependencies with retry logic"""
    # Use uv if available, otherwise fall back to pip
    try:
        cmd = ["uv", "sync", "--upgrade"]
        log.debug("Executing uv sync --upgrade in %s" % cwd)
    except FileNotFoundError:
        cmd = ["pip", "install", "--upgrade", "-e", "."]
        log.debug("Executing pip install --upgrade in %s" % cwd)

    p = subprocess.Popen(cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    res = p.wait()
    out, err = p.communicate()

    if res != 0:
        raise RuntimeError(f"Command failed with exit code {res}: {err.decode()}")

    return out.decode(), err.decode()


def command_update(bot, user, channel, args):
    """Update bot sources from git"""
    if not bot.isAdmin(user):
        return

    pull_ok = False
    pip_ok = False
    cwd = sys.path[0]

    # Git pull with GitPython and retry logic
    try:
        log.debug("Executing git pull in %s" % cwd)
        pull_info = _git_pull(cwd)

        if pull_info:
            pull_ok = True
            bot.say(channel, "Git update OK:")
            for info in pull_info:
                if hasattr(info, "commit"):
                    bot.say(
                        channel,
                        f"Updated to {info.commit.hexsha[:8]}: {info.commit.summary}",
                    )
        else:
            pull_ok = True
            bot.say(channel, "Already up to date")

    except GitCommandError as e:
        bot.say(channel, f"Git pull failed: {e}")
        log.error(f"Git pull failed: {e}")
    except Exception as e:
        bot.say(channel, f"Git operation failed: {e}")
        log.error(f"Git operation failed: {e}")

    # Install/upgrade dependencies
    try:
        out, err = _install_dependencies(cwd)
        bot.say(channel, "Package dependencies updated successfully")
        pip_ok = True

        # Log detailed output for debugging
        if out:
            log.debug(f"Package install output: {out}")
        if err:
            log.debug(f"Package install stderr: {err}")

    except Exception as e:
        bot.say(channel, f"Package update failed: {e}")
        log.error(f"Package update failed: {e}")

    # Rehash after successful update
    if pip_ok and pull_ok:
        bot.command_rehash(user, channel, args)
